# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.subroutines import len_or_one
from probeye.subroutines import make_list
from probeye.subroutines import underlined_string
from probeye.subroutines import titled_table
from probeye.subroutines import replace_string_chars
from probeye.subroutines import simplified_list_string
from probeye.subroutines import simplified_dict_string
from probeye.subroutines import unvectorize_dict_values
from probeye.subroutines import sub_when_empty
from probeye.subroutines import dict2list
from probeye.subroutines import list2dict
from probeye.subroutines import flatten
from probeye.subroutines import process_spatial_coordinates
from probeye.subroutines import translate_prms_def

class TestProblem(unittest.TestCase):

    def test_len_or_one(self):
        # check for objects with no len-method
        self.assertEqual(len_or_one(1), 1)
        self.assertEqual(len_or_one(1.0), 1)
        self.assertEqual(len_or_one(True), 1)
        self.assertEqual(len_or_one("a"), 1)
        self.assertEqual(len_or_one(None), 1)
        # check for lists with zero or more elements
        self.assertEqual(len_or_one([]), 0)
        self.assertEqual(len_or_one([1]), 1)
        self.assertEqual(len_or_one([1, 2]), 2)
        self.assertEqual(len_or_one([1, 2, 3]), 3)
        # check for the weird numpy case (the object np.array(1.0) does have
        # len-method, but when you do len(np.array(1.0)) you get an error)
        self.assertEqual(len_or_one(np.array(1.0)), 1)

    def test_make_list(self):
        # check main use for single non-list input
        self.assertEqual(make_list('a'), ['a'])
        # check common use for lists
        self.assertEqual(make_list(['a']), ['a'])
        self.assertEqual(make_list(['a', 'b']), ['a', 'b'])
        self.assertEqual(make_list([]), [])
        # check common use for tuples
        self.assertEqual(make_list(('a', 'b')), ('a', 'b'))
        self.assertEqual(make_list(()), ())

    def test_underlined_string(self):
        # check common use cases
        self.assertEqual(underlined_string("Test"), "\nTest\n====\n")
        self.assertEqual(underlined_string(
            "ABC", n_empty_start=0, n_empty_end=2, symbol="-"), "ABC\n---\n\n")

    def test_titled_table(self):
        # check common use case
        titled_table_str = titled_table(
            'Title', 'Col Col', symbol='=', n_empty_start=2, n_empty_end=1)
        self.assertEqual(titled_table_str, "\n\nTitle\n=======\nCol Col\n\n")

    def test_replace_string_chars(self):
        # check common use case
        self.assertEqual(replace_string_chars(
            'Test1233', replace={'T': 't', '3': '4'}), 'test1244')
        self.assertEqual(replace_string_chars(
            'Test1233', remove=['1', 'e']), 'Tst233')
        self.assertEqual(replace_string_chars(
            'Test1233', replace={'3': '4'}, remove=['1']), 'Test244')

    def test_simplified_list_string(self):
        # check common use case
        self.assertEqual(simplified_list_string(['a', 'b']), 'a, b')
        self.assertEqual(simplified_list_string(['a']), 'a')

    def test_simplified_dict_string(self):
        # check common use case
        self.assertEqual(simplified_dict_string({'a': 1, 'b': 2}), 'a=1, b=2')
        self.assertEqual(simplified_dict_string({'a': 1.0}), 'a=1.0')

    def test_unvectorize_dict_values(self):
        # check common use case
        self.assertEqual(unvectorize_dict_values({'x': [1, 2], 'y': [3, 4]}),
                         [{'x': 1, 'y': 3}, {'x': 2, 'y': 4}])
        # check for cases with invalid input
        with self.assertRaises(RuntimeError):
            # values cannot have different lengths
            unvectorize_dict_values({'x': [1, 2, 3], 'y': [3, 4]})

    def test_sub_when_empty(self):
        # check common use cases
        self.assertEqual(sub_when_empty('test'), 'test')
        self.assertEqual(sub_when_empty('', empty_str='-'), '-')
        # check for cases with invalid input
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            sub_when_empty(1.0)

    def test_dict2list(self):
        # check common use cases
        self.assertEqual(dict2list({'a': 1, 'b': 2}), [{'a': 1}, {'b': 2}])
        self.assertEqual(dict2list({'a': 1}), [{'a': 1}])
        # check for cases with invalid input
        with self.assertRaises(TypeError):
            # noinspection PyTypeChecker
            dict2list([{'a': 1}, {'b': 2}])

    def test_list2dict(self):
        # check common use cases
        self.assertEqual(list2dict(['a', 'b']),
                         {'a': 'a', 'b': 'b'})
        self.assertEqual(list2dict([{'a': 'm'}, 'b']),
                         {'a': 'm', 'b': 'b'})
        self.assertEqual(list2dict([{'a': 'm'}, {'b': 'c'}]),
                         {'a': 'm', 'b': 'c'})
        self.assertEqual(list2dict({'a': 'm', 'b': 'c'}),
                         {'a': 'm', 'b': 'c'})
        # each dict in the list must not contain more than one element
        with self.assertRaises(ValueError):
            list2dict([{'a': 'm1', 'b': 'b'}])
        # the input cannot contain numbers
        with self.assertRaises(TypeError):
            list2dict(['a', 1.0])

    def test_flatten(self):
        # test for a simple nested list
        nested_list_1 = [1, 2, [3, 4]]
        computed_result = flatten(nested_list_1)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for another simple nested list
        nested_list_2 = [[1, 2, [3, 4]]]
        computed_result = flatten(nested_list_2)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for a flat list
        flat_list = [1, 2, 3, 4]
        computed_result = flatten(flat_list)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for a simple nested numpy-array
        nested_array_1 = np.array([[1, 2], [3, 4]])
        computed_result = flatten(nested_array_1)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for another simple nested numpy-array
        nested_array_2 = np.array([[1, 2, 3, 4]])
        computed_result = flatten(nested_array_2)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for a flat numpy-array
        flat_array = np.array([1, 2, 3, 4])
        computed_result = flatten(flat_array)
        expected_result = [1, 2, 3, 4]
        self.assertEqual(computed_result, expected_result)
        # test for use with scalar
        scalar = 1.2
        computed_result = flatten(scalar)
        expected_result = [1.2]
        self.assertEqual(computed_result, expected_result)
        # test for the None-input
        none_arg = None
        computed_result = flatten(none_arg)
        expected_result = None
        self.assertEqual(computed_result, expected_result)

    def test_process_spatial_coordinates(self):
        # check for complete None-input
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=None, z=None, coords=None)
        coords_expected = np.array([])
        order_expected = []
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for only x being given as numpy array
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=None, z=None, coords=None)
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ['x']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for only x being given as list
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=None, z=None, coords=None)
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ['x']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for only y being given as numpy array
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=np.array([1, 2, 3]), z=None, coords=None)
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ['y']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for only y being given as list
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=[1, 2, 3], z=None, coords=None)
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ['y']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for only z being given as numpy array
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=None, z=np.array([1, 2, 3]), coords=None)
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ['z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for only z being given as list
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=None, z=[1, 2, 3], coords=None)
        coords_expected = np.array([[1, 2, 3]])
        order_expected = ['z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for x and y being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=np.array([4, 5, 6]), z=None, coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ['x', 'y']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for x and y being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=[4, 5, 6], z=None, coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ['x', 'y']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for x and z being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=None, z=np.array([4, 5, 6]), coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ['x', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for x and z being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=None, z=[4, 5, 6], coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ['x', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for y and z being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=np.array([1, 2, 3]), z=np.array([4, 5, 6]), coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ['y', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for y and z being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=None, y=[1, 2, 3], z=[4, 5, 6], coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6]])
        order_expected = ['y', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for x, y and z being given as numpy arrays
        coords_computed, order_computed = process_spatial_coordinates(
            x=np.array([1, 2, 3]), y=np.array([4, 5, 6]),
            z=np.array([7, 8, 9]), coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        order_expected = ['x', 'y', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for y and z being given as lists
        coords_computed, order_computed = process_spatial_coordinates(
            x=[1, 2, 3], y=[4, 5, 6], z=[7, 8, 9], coords=None)
        coords_expected = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        order_expected = ['x', 'y', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)
        # check for coords being given as numpy array with new order
        coords_computed, order_computed = process_spatial_coordinates(
            coords=np.array([[4, 5, 6], [1, 2, 3], [7, 8, 9]]),
            order=('y', 'x', 'z'))
        coords_expected = np.array([[4, 5, 6], [1, 2, 3], [7, 8, 9]])
        order_expected = ['y', 'x', 'z']
        self.assertTrue(np.allclose(coords_computed, coords_expected) and
                        coords_computed.shape == coords_expected.shape)
        self.assertEqual(order_computed, order_expected)

    def test_translate_prms_def(self):
        # valid use case: single string
        prms_def_given = 'sigma'
        expected_result = {'sigma': 'sigma'}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: single string as list
        prms_def_given = ['sigma']
        expected_result = {'sigma': 'sigma'}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: single 1-element dict
        prms_def_given = {'sigma': 'sigma'}
        expected_result = {'sigma': 'sigma'}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: multiple-element dict
        prms_def_given = {'sigma': 'std', 'mu': 'mean'}
        expected_result = {'sigma': 'std', 'mu': 'mean'}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: multiple-element list of strings
        prms_def_given = ['sigma', 'mu']
        expected_result = {'sigma': 'sigma', 'mu': 'mu'}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # valid use case: mixed multiple-element list
        prms_def_given = ['sigma', {'mu': 'mean'}]
        expected_result = {'sigma': 'sigma', 'mu': 'mean'}
        computed_result, _ = translate_prms_def(prms_def_given)
        self.assertEqual(computed_result, expected_result)
        # invalid use case: list with 2-element dict
        with self.assertRaises(ValueError):
            translate_prms_def([{'sigma': 'sigma', 'mu': 'mean'}])

if __name__ == "__main__":
    unittest.main()



# standard library imports
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.sensor import Sensor


class TestProblem(unittest.TestCase):

    def test_sensor_no_coords_init(self):
        # so far the Sensor class has only a name attribute, no methods
        sensor_name = 'Some sensor-name'
        sensor = Sensor(sensor_name)
        # check if everything is where it should be
        self.assertEqual(sensor.name, sensor_name)

    def test_sensor_with_coords_init(self):
        # check a position sensor with three coordinates via x-y-z-input
        x, y, z, sensor_name = 1, 2, 3, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, x=x, y=y, z=z)
        # check if everything is where it should be
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.coords.shape, (3, 1))
        self.assertEqual(position_sensor.order, ['x', 'y', 'z'])
        for i, value in enumerate([x, y, z]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with three coordinates via coords-input
        coords, sensor_name = np.array([[1], [2], [3]]), 'Some sensor-name'
        position_sensor = Sensor(sensor_name, coords=coords)
        # check if everything is where it should be
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.coords.shape, (3, 1))
        self.assertEqual(position_sensor.order, ['x', 'y', 'z'])
        for i, value in enumerate([x, y, z]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with two coordinates (x and y) via x-y-input
        x, y, sensor_name = 1, 2, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, x=x, y=y)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.coords.shape, (2, 1))
        self.assertEqual(position_sensor.order, ['x', 'y'])
        for i, value in enumerate([x, y]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with two coordinates (x, y) via coords-input
        coords, sensor_name = np.array([[1], [2]]), 'Some sensor-name'
        position_sensor = Sensor(sensor_name, coords=coords)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.order, ['x', 'y'])
        self.assertEqual(position_sensor.coords.shape, (2, 1))
        for i, value in enumerate([x, y]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with two coordinates (x and z) via x-z-input
        x, z, sensor_name = 1, 3, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, x=x, z=z)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.order, ['x', 'z'])
        self.assertEqual(position_sensor.coords.shape, (2, 1))
        for i, value in enumerate([x, z]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with two coordinates (x, z) via coords-input
        coords, sensor_name = np.array([[1], [3]]), 'Some sensor-name'
        position_sensor = Sensor(
            sensor_name, coords=coords, order=('x', 'z'))
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.order, ['x', 'z'])
        self.assertEqual(position_sensor.coords.shape, (2, 1))
        for i, value in enumerate([x, z]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with two coordinates (y and z) via y-z-input
        y, z, sensor_name = 2, 3, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, y=y, z=z)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.order, ['y', 'z'])
        self.assertEqual(position_sensor.coords.shape, (2, 1))
        for i, value in enumerate([y, z]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with two coordinates (y, z) via coords-input
        coords, sensor_name = np.array([[2], [3]]), 'Some sensor-name'
        position_sensor = Sensor(
            sensor_name, coords=coords, order=('y', 'z'))
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.order, ['y', 'z'])
        self.assertEqual(position_sensor.coords.shape, (2, 1))
        for i, value in enumerate([y, z]):
            self.assertAlmostEqual(position_sensor.coords[i], value)

        # check a position sensor with one coordinate (x) via x-input
        x, sensor_name = 1, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, x=x)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, x)
        self.assertEqual(position_sensor.order, ['x'])
        self.assertEqual(position_sensor.coords.shape, (1, 1))
        self.assertAlmostEqual(position_sensor.coords[0], x)

        # check a position sensor with one coordinate (x) via coords-input
        coords, sensor_name = np.array([[1]]), 'Some sensor-name'
        position_sensor = Sensor(sensor_name, coords=coords)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.x, 1)
        self.assertEqual(position_sensor.order, ['x'])
        self.assertEqual(position_sensor.coords.shape, (1, 1))
        self.assertAlmostEqual(position_sensor.coords[0], 1)

        # check a position sensor with one coordinate (y) via y-input
        y, sensor_name = 2, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, y=y)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.y, y)
        self.assertEqual(position_sensor.order, ['y'])
        self.assertEqual(position_sensor.coords.shape, (1, 1))
        self.assertAlmostEqual(position_sensor.coords[0], y)

        # check a position sensor with one coordinate (y) via coords-input
        coords, sensor_name = np.array([[2]]), 'Some sensor-name'
        position_sensor = Sensor(
            sensor_name, coords=coords, order='y')
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.y, 2)
        self.assertEqual(position_sensor.order, ['y'])
        self.assertEqual(position_sensor.coords.shape, (1, 1))
        self.assertAlmostEqual(position_sensor.coords[0], 2)

        # check a position sensor with one coordinate (z) via z-input
        z, sensor_name = 3, 'Some sensor-name'
        position_sensor = Sensor(sensor_name, z=z)
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.z, z)
        self.assertEqual(position_sensor.order, ['z'])
        self.assertEqual(position_sensor.coords.shape, (1, 1))
        self.assertAlmostEqual(position_sensor.coords[0], z)

        # check a position sensor with one coordinate (z) via coords-input
        coords, sensor_name = np.array([[3]]), 'Some sensor-name'
        position_sensor = Sensor(
            sensor_name, coords=coords, order='z')
        self.assertEqual(position_sensor.name, sensor_name)
        self.assertEqual(position_sensor.z, 3)
        self.assertEqual(position_sensor.order, ['z'])
        self.assertEqual(position_sensor.coords.shape, (1, 1))
        self.assertAlmostEqual(position_sensor.coords[0], 3)

if __name__ == "__main__":
    unittest.main()



# third party imports
import numpy as np

# local imports
from probeye.definition.noise_model import NormalNoiseModel
from probeye.inference.scipy_.correlation_models import \
    SpatialExponentialCorrelationModel


class NormalNoise(NormalNoiseModel):

    def __init__(self, target_sensor, prms_def, name=None, corr=None,
                 corr_model='exp', noise_type='additive'):
        """
        Parameters
        ----------
        target_sensor : obj[Sensor]
            The sensor, the noise refers to. Note that this is not a sensor's
            name but a Sensor object. This object must contain the spatial
            information that is referred to in corr. For example, if corr is set
            to 'xy' then target_sensor must have a x and a y attribute that
            contain coordinate vectors of similar lengths.
        prms_def : {str, list, dict}
            Contains the noise model's parameter names. The list may only
            contain strings or one-element dictionaries. It could look, for
            example, like [{'a': 'm'}, 'b']. The one-element dictionaries
            account for the possibility to define a local name for a latent
            parameter that is different from the global name. In the example
            above, the latent parameter with the global name 'a' will be
            referred to as 'm' within the noise model. So, the one-element dicts
            have the meaning {<global name>: <local name>}. String-elements are
            interpreted as having similar local and global names. Note that the
            local-name option will not be required most of the times. The input
            from global to local name can also be provided as a dict. In the
            example above it would look like {'a': 'm', 'b': 'b'}.
        name : {str, None}, optional
            Unique name of the noise model. This name is None, if the user does
            not specify it when adding the noise model to the problem. It is
            then named automatically before starting the inference engine.
        corr : {'x', 'y', 'z', 'xy', 'xz', 'yz', 'xyz', None}, optional
            Spatial coordinates to be considered when setting up the covariance
            matrix. For example, when 'x' is chosen, the distance between two
            points i and j is computed as |xi - xj|. If 'xz' is chosen, this
            distance would be computed as sqrt((xi-xj)**2 + (zi-zj)**2). If None
            is chosen, no spatial correlation will be considered.
        corr_model : {'exp'}, optional
            Defines the correlation function the covariance matrix will be based
            on. Currently, there is only the one optional of an exponential
            model ('exp'). More might be added in future versions. This argument
            has no effect if corr is set to None.
        noise_type : {'additive', 'multiplicative', 'other'}, optional
            Defines if the model error is computed by [prediction - measurement]
            ('additive') or via [prediction/measurement-1] ('multiplicative') or
            in some 'other' i.e., non-standard fashion.
        """

        # initialize the super-class (NormalNoiseModel) based on the given input
        super().__init__(prms_def=prms_def, sensors=target_sensor, name=name,
                         corr=corr, corr_model=corr_model,
                         noise_type=noise_type)

        # the target_sensor is not an attribute of the super-class
        self.target_sensor = target_sensor

        # correlation related attributes
        self.corr = corr
        self.corr_model = corr_model
        if self.corr is not None:
            if self.corr_model == 'exp':
                self.cov = SpatialExponentialCorrelationModel(
                    coords=target_sensor.coords, order=list(self.corr))
                self.loglike_contribution = \
                    self.loglike_contribution_with_correlation
            else:
                raise ValueError(
                    f"Encountered unknown flag '{self.corr_model}' "
                    f"for requested correlation model.\n Currently available "
                    f"options are: 'exp' for an exponential model.")
        else:
            self.loglike_contribution =\
                self.loglike_contribution_without_correlation

    def loglike_contribution_without_correlation(self, model_response, prms):
        """
        This method overwrites the corresponding method of the parent class.
        Check out the docstring there for additional information.
        """
        # compute the model error; note that this mode has exactly one sensor
        model_error_vector = self.error(model_response)[self.target_sensor.name]
        # the precision 'prec' is defined as the inverse of the variance, hence
        # prec = 1 / sigma**2 where sigma denotes the standard deviation
        std = prms['std']
        mean = 0.0 if self.zero_mean else prms['mean']
        prec = 1.0 / std ** 2.0
        # evaluate the Gaussian log-PDF with zero mean and a variance of
        # 1/prec for each error term and sum them up
        ll = -len(model_error_vector) / 2 * np.log(2 * np.pi / prec)
        ll -= 0.5 * prec * np.sum(np.square(model_error_vector - mean))
        return ll

    def loglike_contribution_with_correlation(self, model_response_dict, prms,
                                              worst_value=-np.infty):
        """
        Evaluates the log-likelihood for the noise model's experiments.

        Parameters
        ----------
        model_response_dict : dict
            The first key is the name of the experiment. The values are dicts
            which contain the forward model's output sensor's names as keys
            have the corresponding model responses as values.
        prms : dict
            Contains the names of the correlation model's parameters as keys
            and the corresponding numeric values as values.
        worst_value : float, int, optional
            This value is returned when this method does not result in a numeric
            value. This might happen for example when the given parameters are
            not valid (for example in case of a negative standard deviation).
            The returned value in such cases should represent the worst possible
            value of the contribution.

        Returns
        -------
        ll : float
            The evaluated log-likelihood function.
        """
        # before computing the covariance matrix, check if the given parameter
        # values are valid; if not, the computation of the log-likelihood
        # contribution is stopped right away while returning the worst value
        # its evaluation can result in
        if not self.cov.check_prms(prms):
            return worst_value
        # since the covariance matrix is the same for all experiments in the
        # noise model, values derived from it (its inverse and determinant) are
        # computed before entering the loop over the noise model's experiments
        cov_matrix = self.cov(prms)
        inv_cov_matrix = np.linalg.inv(cov_matrix)
        # the following command computes the natural log and the sign of the
        # matrix cov_matrix; note that cov_matrix is positive definite, hence
        # the sign will always be positive and therefore isn't requested here
        _, log_det_cov_matrix = np.linalg.slogdet(cov_matrix)
        n = self.cov.n  # number of data points recorded by the sensor
        n_exp = len(self.experiment_names)
        # the term in the first parenthesis is added by each experiment in the
        # loop; this is why we can pre-compute it outside the loop
        ll = n_exp * (-(n * np.log(2 * np.pi) + log_det_cov_matrix) / 2)
        for exp_name in self.experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            ym = model_response_dict[exp_name]
            ye = exp_dict['sensor_values']
            error = self.error_function(ym, ye)[self.target_sensor.name]
            ll += -np.dot(error, inv_cov_matrix.dot(error)) / 2
        return ll

def translate_noise_model(noise_base):
    """
    Translates a given instance of NoiseBase (which is essentially just a
    description of the noise model without computing-methods) to a specific
    noise model object which does contain computing-methods (e.g. compute the
    log-likelihood contribution).

    Parameters
    ----------
    noise_base : obj[NoiseBase]
        An instance of NoiseBase which contains basic information on the noise
        model but no computing-methods.

    Returns
    -------
    noise_object : obj[NoiseBase]
        An instance of a specific noise model class with computing capabilities.
        Examples for such classes are given above in this file.
    """

    # this is the noise classes currently defined (see code above)
    noise_classes = {'normal': NormalNoise}

    # this is where the translation happens
    noise_object = noise_classes[noise_base.dist](
        target_sensor=noise_base.sensors[0], prms_def=noise_base.prms_def,
        name=noise_base.name, corr=noise_base.corr,
        corr_model=noise_base.corr_model, noise_type=noise_base.noise_type)

    # here, we take the assigned experiments from the base object
    noise_object.experiment_names = noise_base.experiment_names
    noise_object.problem_experiments = noise_base.problem_experiments

    return noise_object



# standard library imports
import unittest

# local imports
from probeye.definition.parameter import Parameters
from probeye.definition.parameter import ParameterProperties
from probeye.definition.prior import PriorBase


class TestProblem(unittest.TestCase):

    def test_parameters(self):
        # this is the standard way to use Parameters
        parameters = Parameters()
        parameters['a'] = ParameterProperties({'index': None,
                                               'type': 'model',
                                               'prior': None,
                                               'value': 1.0,
                                               'info': "...",
                                               'tex': r"$a$"})
        parameters['b'] = ParameterProperties({'index': None,
                                               'type': 'model',
                                               'prior': None,
                                               'value': 2.0,
                                               'info': "...",
                                               'tex': r"$b$"})
        c_prior = PriorBase('c', ['s'], 'c_dummy', 'normal')
        parameters['c'] = ParameterProperties({'index': 2,
                                               'type': 'model',
                                               'prior': c_prior,
                                               'value': None,
                                               'info': "...",
                                               'tex': r"$c$"})
        # check if the role is correctly identified
        self.assertEqual(parameters['a'].role, 'const')
        self.assertEqual(parameters['c'].role, 'latent')
        # check for the parameter-role bookkeeping
        self.assertEqual(parameters.n_latent_prms, 1)
        self.assertEqual(parameters.n_constant_prms, 2)
        self.assertEqual(parameters.n_prms, 3)
        # you cannot add the parameter's properties as a dictionary
        with self.assertRaises(ValueError):
            d_prior = PriorBase('d', ['s'], 'd_dummy', 'normal')
            parameters['d'] = {'index': 3,
                               'type': 'model',
                               'role': 'latent',
                               'prior': d_prior,
                               'value': None,
                               'info': "...",
                               'tex': r"$d$"}

        # you should not be able to set parameter attributes directly
        with self.assertRaises(AttributeError):
            parameters['a'].index = -1
        with self.assertRaises(AttributeError):
            parameters['a'].type = -1
        with self.assertRaises(AttributeError):
            parameters['a'].role = -1
        with self.assertRaises(AttributeError):
            parameters['a'].prior = -1
        with self.assertRaises(AttributeError):
            parameters['a'].value = -1

        # check if the consistency-checks work
        with self.assertRaises(TypeError):
            # index is has wrong type
            parameters['d'] = ParameterProperties({'index': True,
                                                   'type': 'model',
                                                   'prior': None,
                                                   'value': None,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        with self.assertRaises(RuntimeError):
            # index has invalid value
            parameters['d'] = ParameterProperties({'index': -1,
                                                   'type': 'model',
                                                   'prior': None,
                                                   'value': 1.0,
                                                   'info': "...",
                                                   'tex': r"$d$"})

        with self.assertRaises(RuntimeError):
            # type has invalid value
            parameters['d'] = ParameterProperties({'index': None,
                                                   'type': 'invalid value',
                                                   'prior': None,
                                                   'value': 1.0,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        with self.assertRaises(TypeError):
            # prior has invalid type
            parameters['d'] = ParameterProperties({'index': None,
                                                   'type': 'model',
                                                   'prior': True,
                                                   'value': 1.0,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        with self.assertRaises(TypeError):
            # value has invalid type
            parameters['d'] = ParameterProperties({'index': None,
                                                   'type': 'model',
                                                   'prior': None,
                                                   'value': True,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        
        with self.assertRaises(RuntimeError):
            # index and prior are inconsistently combined
            parameters['d'] = ParameterProperties({'index': 3,
                                                   'type': 'model',
                                                   'prior': None,
                                                   'value': None,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        with self.assertRaises(RuntimeError):
            # index and value are inconsistently combined
            d_prior = PriorBase('d', ['s'], 'd_dummy', 'normal')
            parameters['d'] = ParameterProperties({'index': 3,
                                                   'type': 'model',
                                                   'prior': d_prior,
                                                   'value': 1.0,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        with self.assertRaises(RuntimeError):
            # index and prior are inconsistently combined
            d_prior = PriorBase('d', ['s'], 'd_dummy', 'normal')
            parameters['d'] = ParameterProperties({'index': None,
                                                   'type': 'model',
                                                   'prior': d_prior,
                                                   'value': None,
                                                   'info': "...",
                                                   'tex': r"$d$"})
        with self.assertRaises(RuntimeError):
            # index and value are inconsistently combined
            parameters['d'] = ParameterProperties({'index': None,
                                                   'type': 'model',
                                                   'prior': None,
                                                   'value': None,
                                                   'info': "...",
                                                   'tex': r"$d$"})

if __name__ == "__main__":
    unittest.main()



# third party imports
import pyro
import pyro.distributions as dist

# local imports
from probeye.definition.prior import PriorBase


class PriorNormal(PriorBase):
    """Prior class for a normal distribution."""

    def __init__(self, ref_prm, prms_def, name):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings defining the prior's parameter names.
        name : string
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "normal distribution")

    def sample(self, prms):
        """
        Evaluates stats.norm.<method>(x, loc, scale). This function is mostly
        used with method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.

        Returns
        -------
        float
            The result of stats.norm.<method>(x, loc, scale).
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        return pyro.sample(self.ref_prm, dist.Normal(loc, scale))


class PriorUniform(PriorBase):
    """Prior class for a uniform distribution."""

    def __init__(self, ref_prm, prms_def, name):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings defining the prior's parameter names.
        name : string
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "uniform distribution")

    def sample(self, prms):
        """
        Evaluates stats.uniform.<method>(x, loc, scale). This function is mostly
        used with method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.

        Returns
        -------
        float
            The result of stats.uniform.<method>(x, loc, scale).
        """
        low = prms[f"low_{self.ref_prm}"]
        high = prms[f"high_{self.ref_prm}"]
        return pyro.sample(self.ref_prm, dist.Uniform(low, high))

def translate_prior_template(prior_template):
    """
    Translate a given instance of PriorBase (which is essentially just a
    description of the prior without compute-methods) to a specific prior object
    which does contain compute-methods (e.g. compute the logpdf).

    Parameters
    ----------
    prior_template : obj[PriorBase]
        An instance of PriorBase which contains basic information on the
        prior but no computing-methods.

    Returns
    -------
    prior_object : obj[PriorBase]
        An instance of a specific prior class with computing capabilities.
        Examples for such classes are given above in this file.
    """

    # this dictionary contains the classes from above in this file
    prior_classes = {'normal': PriorNormal,
                     'uniform': PriorUniform}

    # prepare the corresponding prior object; the following translation is
    # necessary, because prms_def must be given in form of a list, but was
    # already translated to a dictionary when instantiating the PriorBase
    # objects; hence prior_template.prms_def is a dictionary
    prms_def = [{key: value} for key, value in prior_template.prms_def.items()
                if key != prior_template.ref_prm]
    prior_object = prior_classes[prior_template.prior_type](
        prior_template.ref_prm, prms_def, prior_template.name)

    return prior_object



# standard library imports
import unittest

# third party imports
from scipy import stats

# local imports
from probeye.inference.scipy_.priors import PriorNormal
from probeye.inference.scipy_.priors import PriorLognormal
from probeye.inference.scipy_.priors import PriorUniform
from probeye.inference.scipy_.priors import PriorWeibull


class TestProblem(unittest.TestCase):

    def test_prior_normal(self):
        prior_normal = PriorNormal('a', ['loc_a', 'scale_a'], 'a_normal')
        # check the evaluation of the log-pdf
        prms = {'a': 1.0, 'loc_a': 0.0, 'scale_a': 1.0}
        self.assertEqual(
            stats.norm.logpdf(prms['a'], prms['loc_a'], prms['scale_a']),
            prior_normal(prms, 'logpdf'))
        # check the sampling-method (samples are checked one by one)
        prms = {'loc_a': 0.0, 'scale_a': 1.0}
        prior_samples = prior_normal.generate_samples(prms, 10, seed=1)
        sp_samples = stats.norm.rvs(loc=prms['loc_a'], scale=prms['scale_a'],
                                    size=10, random_state=1)
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

    def test_prior_lognormal(self):
        prior_lognormal = PriorLognormal(
            'a', ['loc_a', 'scale_a'], 'a_lognormal')
        # check the evaluation of the log-pdf
        prms = {'a': 2.0, 'loc_a': 1.0, 'scale_a': 1.0}
        self.assertEqual(
            stats.lognorm.logpdf(prms['a'], prms['loc_a'], prms['scale_a']),
            prior_lognormal(prms, 'logpdf'))
        # check the sampling-method (samples are checked one by one)
        prms = {'loc_a': 1.0, 'scale_a': 1.0}
        prior_samples = prior_lognormal.generate_samples(prms, 10, seed=1)
        sp_samples = stats.lognorm.rvs(1.0,  # this is scipy's shape parameter
                                       loc=prms['loc_a'], scale=prms['scale_a'],
                                       size=10, random_state=1)
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

    def test_prior_uniform(self):
        prior_uniform = PriorUniform('a', ['low_a', 'high_a'], 'a_uniform')
        # check the evaluation of the log-pdf
        prms = {'a': 0.5, 'low_a': 0.0, 'high_a': 1.0}
        self.assertEqual(
            stats.uniform.logpdf(prms['a'], prms['low_a'], prms['high_a']),
            prior_uniform(prms, 'logpdf'))
        # check the sampling-method (samples are checked one by one)
        prms = {'low_a': 0.0, 'high_a': 1.0}
        prior_samples = prior_uniform.generate_samples(prms, 10, seed=1)
        sp_samples = stats.uniform.rvs(
            loc=prms['low_a'], scale=prms['low_a'] + prms['high_a'],
            size=10, random_state=1)
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

    def test_prior_weibull(self):
        prior_weibull = PriorWeibull(
            'a', ['loc_a', 'scale_a', 'shape_a'], 'a_weibull')
        # check the evaluation of the log-pdf
        prms = {'a': 1.0, 'loc_a': 1.0, 'scale_a': 1.0, 'shape_a': 2.0}
        self.assertEqual(
            stats.weibull_min.logpdf(prms['a'], prms['shape_a'],
                                     prms['loc_a'], prms['scale_a']),
            prior_weibull(prms, 'logpdf'))
        # check the sampling-method (samples are checked one by one)
        prms = {'loc_a': 1.0, 'scale_a': 1.0, 'shape_a': 2.0}
        prior_samples = prior_weibull.generate_samples(prms, 10, seed=1)
        sp_samples = stats.weibull_min.rvs(
            prms['shape_a'], loc=prms['loc_a'], scale=prms['scale_a'],
            size=10, random_state=1)
        for s1, s2 in zip(prior_samples, sp_samples):
            self.assertEqual(s1, s2)

if __name__ == "__main__":
    unittest.main()



# third party imports
from tabulate import tabulate
from loguru import logger

# local imports
from probeye.subroutines import titled_table
from probeye.subroutines import simplified_list_string
from probeye.definition.prior import PriorBase

class Parameters(dict):
    """
    The main parameter 'library'. In this dictionary, all of the problem's
    parameters are stored. The parameter's names are the keys, and the
    associated values are ParameterProperties-objects, see below.
    """

    def add_parameter(self, prm_name, prm_type, const=None, prior=None,
                      info="No explanation provided", tex=None):
        """
        Adds a parameter ('const' or 'latent') to the Parameters-object. The
        main functionality of this method is to distinguish between the two
        types ('const' and 'latent') and in creating the prior-object and
        adding the prior-parameters when adding a latent param. to the problem.

        Parameters
        ----------
        prm_name : str
            The name of the parameter which should be added to the problem.
        prm_type : str
            Either 'model' (for a model parameter), 'prior' (for a prior
            parameter) or 'noise' (for a noise parameter).
        const : float or None, optional
            If the added parameter is a 'const'-parameter, the corresponding
            value has to be specified by this argument.
        prior : tuple or list of two elements or None, optional
            If the added parameter is a 'latent'-parameter, this argument
            has to be given as a 2-tuple. The first element (a string) defines
            the prior-type (will be referenced in inference routines). The 2nd
            element must be a dictionary stating the prior's parameters as keys
            and their numeric values as values or the name of a pre-defined
            parameter within the problem scope. An example for a normal prior:
            ('normal', {'loc': 0.0, 'scale': 1.0}). In order to define the
            prior's parameters, check out the prior definitions in priors.py.
        info : str, optional
            Short explanation on the added parameter.
        tex : str or None, optional
            The TeX version of the parameter's name, for example r'$\beta$'
            for a parameter named 'beta'.
        """

        # make sure the parameters has not been defined yet
        self.confirm_that_parameter_does_not_exists(prm_name)

        # if neither const nor prior are given, the parameter is interpreted as
        # being defined as latent with an uninformative prior
        if const is None and prior is None:
            prior = ('uninformative', {})

        # add the parameter to the central parameter dictionary
        if const is None:  # i.e. adding 'latent'-parameter
            # first, define the index of this parameter in the numeric vector
            # theta, which is given to self.loglike and self.logprior
            prm_index = self.n_latent_prms
            # the prm_value is reserved for 'const'-parameter; hence, it is set
            # to None in this case, where we are adding a 'latent'-param.
            prm_value = None
            # the remaining code in this if-branch defines the prior that is
            # associated with this 'latent'-parameter; first, however, check
            # whether the given prior has a valid structure
            if type(prior) not in [list, tuple]:
                raise TypeError(
                    f"The given prior is of type {type(prior)} but must be "
                    f"either a list or a tuple!")
            if len(prior) != 2:
                raise RuntimeError(
                    f"The given prior must be a list/tuple with two elements. "
                    f"However, the given prior has {len(prior)} element(s).")
            if type(prior[0]) is not str:
                raise TypeError(
                    f"The first element of the prior must be of type string. "
                    f"However, the given first element is of type "
                    f"{type(prior[0])}.")
            if type(prior[1]) is not dict:
                raise TypeError(
                    f"The second element of the prior must be of type dict. "
                    f"However, the given second element is of type "
                    f"{type(prior[1])}.")
            # extract the prior's elements
            prior_type = prior[0]  # e.g. 'normal', 'lognormal', etc.
            prior_dict = prior[1]  # dictionary with parameter-value pairs
            prior_parameter_names = []
            for prior_parameter_name, value in prior_dict.items():
                # create unique name for this prior parameter
                new_name = f"{prior_parameter_name}_{prm_name}"
                prior_parameter_names.append(new_name)
                if type(value) in {float, int}:
                    # in this case, the prior-parameter is considered a 'const'-
                    # parameter and added to the problem accordingly here
                    default_info = f"{prior_type.capitalize()} "
                    default_info += f"prior's parameter "
                    default_info += f"for latent parameter '{prm_name}'"
                    # the following call is recursive, but only with a depth of
                    # one, since the added parameter is a constant here
                    self.add_parameter(new_name, 'prior', const=value,
                                       info=default_info)
                elif type(value) is str:
                    # in this case the prior-parameter is defined as an already
                    # defined parameter with the name stated in value
                    self.confirm_that_parameter_exists(value)
                else:
                    raise TypeError(
                        f"The prior-parameter {new_name} is not assigned a "
                        f"float, int or str, but something of type "
                        f"{type(value)}.")
            prior_name = f"{prm_name}_{prior_type}"  # unique name of this prior
            prm_prior = PriorBase(prm_name, prior_parameter_names,
                                  prior_name, prior_type)
            logger.debug(f"Adding  latent  {prm_type}-parameter "
                         f"{prm_name} with {prior_type} prior to problem")

        else:
            # in this case we are adding a 'const'-parameter, which means that
            # the prm_index and prm_prior values are not used here
            prm_index = None
            prm_prior = None
            prm_value = const
            logger.debug(f"Adding constant {prm_type}-parameter "
                         f"{prm_name} = {prm_value} to problem")

        # add the parameter to the central parameter dictionary
        self[prm_name] = ParameterProperties({'index': prm_index,
                                              'type': prm_type,
                                              'prior': prm_prior,
                                              'value': prm_value,
                                              'info': info,
                                              'tex': tex})

    def __setitem__(self, key, value):
        """Performs some checks before adding a parameter to the dictionary."""
        if type(key) != str:
            raise ValueError(
                f"The key must be a parameters name (string), but you provided "
                f"something of type '{type(key)}'.")
        if type(value) != ParameterProperties:
            raise ValueError(
                f"The properties of your parameter must be given in form of an "
                f"ParameterProperties-object. But you provided something of "
                f"type '{type(value)}'.")
        super().__setitem__(key, value)

    def __delitem__(self, key):
        """
        Deletes an item from itself while taking care of additional actions. For
        example removing prior-parameters when deleting a latent parameter or
        keeping the index-attributes of the latent parameters consistent.
        """

        # the given key is a parameter's name
        # (renaming for easier readability)
        prm_name = key

        # check if the given parameter exists
        self.confirm_that_parameter_exists(prm_name)

        # different steps must be taken depending on whether the parameter which
        # should be removed is a 'const'- or a 'latent'-parameter
        if self[prm_name].index is None:
            # in this case prm_name refers to a constant parameter; hence, we
            # can simply remove this parameter without having to take care of
            # other things as we will have to do for latent parameters
            dict.__delitem__(self, key)
        else:
            # in this case prm_name refers to a latent parameter; hence we need
            # to also remove the prior-parameters; also, we have to correct the
            # index values of the remaining latent parameters
            for prior_prm in self[prm_name].prior.hyperparameters.keys():
                self.__delitem__(prior_prm)  # recursive call
            dict.__delitem__(self, prm_name)
            # correct the indices of the remaining 'latent'-parameters; note
            # that the way how the correction is done is due to the fact that
            # the parameter.index attribute is protected, and cannot be changed
            # directly from outside
            idx_dict = {}
            idx = 0
            for prm_name, parameter in self.items():
                if parameter.index is not None:
                    idx_dict[prm_name] = idx
                    idx += 1
            for prm_name, idx in idx_dict.items():
                self[prm_name] = self[prm_name].changed_copy(index=idx)

    def confirm_that_parameter_exists(self, prm_name):
        """
        Checks if a parameter, given by its name, exists among the currently
        defined parameters. An error is raised when the given parameter does not
        exist yet.

        Parameters
        ----------
        prm_name : str
            A global parameter name.
        """
        if prm_name not in self:
            raise RuntimeError(
                f"A parameter with name '{prm_name}' has not been defined yet.")

    def confirm_that_parameter_does_not_exists(self, prm_name):
        """
        Checks if a parameter, given by its name, exists among the currently
        defined parameters. An error is raised when the given parameter does
        already exist.

        Parameters
        ----------
        prm_name : str
            A global parameter name.
        """
        if prm_name in self:
            raise RuntimeError(
                f"A parameter with name '{prm_name}' has already been defined.")

    @property
    def prms(self):
        """Access the names of all parameters as an attribute."""
        return [*self.keys()]

    @property
    def n_prms(self):
        """Access the number of all parameters as an attribute."""
        return len(self)

    @property
    def latent_prms(self):
        """Access the names of all 'latent'-parameters as an attribute."""
        return [name for name, prm in self.items() if prm.is_latent]

    @property
    def n_latent_prms(self):
        """Access the number of all 'latent'-parameters as an attribute."""
        return len(self.latent_prms)

    @property
    def constant_prms(self):
        """Access the names of all 'const'-parameters as an attribute."""
        return [name for name, prm in self.items() if prm.is_const]

    @property
    def constant_prms_dict(self):
        """Access the names and values of all 'const'-param. as an attribute."""
        return {name: prm.value for name, prm in self.items() if prm.is_const}

    @property
    def n_constant_prms(self):
        """Access the number of all 'const'-parameters as an attribute."""
        return len(self.constant_prms)

    @property
    def model_prms(self):
        """Access the names of all 'model'-parameters as an attribute."""
        return [name for name, prm in self.items() if prm.type == "model"]

    @property
    def n_model_prms(self):
        """Access the number of all 'model'-parameters as an attribute."""
        return len(self.model_prms)

    @property
    def prior_prms(self):
        """Access the names of all 'prior'-parameters as an attribute."""
        return [name for name, prm in self.items() if prm.type == "prior"]

    @property
    def n_prior_prms(self):
        """Access the number of all 'prior'-parameters as an attribute."""
        return len(self.prior_prms)

    @property
    def noise_prms(self):
        """Access the names of all 'noise'-parameters as an attribute."""
        return [name for name, prm in self.items() if prm.type == "noise"]

    @property
    def n_noise_prms(self):
        """Access the number of all 'prior'-parameters as an attribute."""
        return len(self.noise_prms)

    def parameter_overview(self, tablefmt="presto"):
        """
        Returns a string providing an overview of the defined parameters.

        Parameters
        ----------
        tablefmt : string, optional
            An argument for the tabulate function defining the style of the
            generated table. Check out tabulate's documentation for more info.

        Returns
        -------
        prm_string : string
            This string describes a nice table with some essential information
            on the parameters of the problem.
        """
        # each element describes one row in the table to be generated
        rows = [('Model parameters',
                simplified_list_string(self.model_prms),
                self.n_model_prms),
                ('Prior parameters',
                simplified_list_string(self.prior_prms),
                self.n_prior_prms),
                ('Noise parameters',
                simplified_list_string(self.noise_prms),
                self.n_noise_prms),
                ('Const parameters',
                simplified_list_string(self.constant_prms),
                self.n_constant_prms),
                ('Latent parameters',
                simplified_list_string(self.latent_prms),
                self.n_latent_prms)]
        # these are the strings appearing in the column headers
        headers = ["Parameter type/role", "Parameter names", "Count"]
        prm_table = tabulate(rows, headers=headers, tablefmt=tablefmt)
        prm_string = titled_table('Parameter overview', prm_table)
        return prm_string

    def parameter_explanations(self, tablefmt="presto"):
        """
        Returns a string providing short explanations on the defined parameters.

        Parameters
        ----------
        tablefmt : string
            An argument for the tabulate function defining the style of the
            generated table. Check out tabulate's documentation for more info.

        Returns
        -------
        prm_string : string, optional
            This string describes a nice table with short explanations on the
            parameters of the problem.
        """
        rows = [(name, prm.info) for name, prm in self.items()]
        headers = ["Name", "Short explanation"]
        prm_table = tabulate(rows, headers=headers, tablefmt=tablefmt)
        prm_string = titled_table('Parameter explanations', prm_table)
        return prm_string

    def const_parameter_values(self, tablefmt="presto"):
        """
        Returns a string providing the values of the defined 'const'-parameters.

        Parameters
        ----------
        tablefmt : string
            An argument for the tabulate function defining the style of the
            generated table. Check out tabulate's documentation for more info.

        Returns
        -------
        prm_string : string, optional
            This string describes a nice table with the names and values of the
            constant parameters of the problem.
        """
        rows = [(name, prm.value)
                for name, prm in self.items() if prm.value is not None]
        headers = ["Name", "Value"]
        prm_table = tabulate(rows, headers=headers, tablefmt=tablefmt)
        prm_string = titled_table('Constant parameters', prm_table)
        return prm_string

class ParameterProperties:
    """
    Describes relevant properties of a ('latent' or 'const') parameter.
    Objects from this class are associated with the parameter's name in the
    dictionary class 'Parameters', see above. The use of this class as opposed
    to a standard dictionary allows convenient auto-completion while coding.
    """

    def __init__(self, prm_dict):
        """
        Parameters
        ----------
        prm_dict : dict
            The keys are 'index', 'type', 'role', 'prior', 'value', 'info' and
            'tex', while the values are the corresponding values of these
            properties. See also the explanations in InferenceProblem.__init__()
            for more detailed information.
        """
        # write attributes
        self._index = prm_dict['index']
        self._type = prm_dict['type']
        self._prior = prm_dict['prior']
        self._value = prm_dict['value']
        self.info = prm_dict['info']
        self.tex = prm_dict['tex']

        # whitespace in the tex strings is a problem for some plotting routines,
        # so they are replaced here by a math-command for whitespace that does
        # not contain actual whitespace
        if self.tex:
            self.tex = self.tex.replace(" ", r'$\enspace$')

        # check the given values
        self.check_consistency()

    # noinspection PyShadowingBuiltins
    def changed_copy(self, index=None, type=None, prior=None, value=None,
                     info=None, tex=None):
        """
        Convenience method that simplifies changing the attributes of a
        ParameterProperties object based on creating a new instance. The reason
        for this approach is that some of the attributes are private, and cannot
        (or at least should not) be changed directly from outside.

        See the explanations in InferenceProblem.__init__() for more detailed 
        information on the arguments.
        """
        return ParameterProperties({
                    "index": index if index is not None else self._index,
                    "type": type or self._type,
                    "prior": prior or self._prior,
                    "value": value or self._value,
                    "info": info or self.info,
                    "tex": tex or self.tex})

    def check_consistency(self):
        """
        Checks the defined attributes in both isolated checks (each attribute
        is checked without considering others) and cross-checks, where the
        combination of attributes is checked on consistency.
        """

        # ------------------------------- #
        #         Isolated checks         #
        # ------------------------------- #

        if not (type(self._index) == int or self._index is None):
            raise TypeError(
                f"Found invalid ParameterProperties._index attribute! It must "
                f"be of type int or None, but found {type(self._index)}.")

        if (self._index is not None) and (self._index < 0):
            raise RuntimeError(
                f"Found negative value for ParameterProperties._index! This "
                f"attribute must be a non-negative integer, but found a value "
                f"of {self._index}.")

        if self._type not in ['model', 'prior', 'noise']:
            raise RuntimeError(
                f"Found invalid ParameterProperties._type attribute! It can "
                f"only assume the three values 'model', 'prior' or 'noise' but "
                f"found '{self._type}'.")

        if not (type(self._prior) == PriorBase or self._prior is None):
            raise TypeError(
                f"Found invalid ParameterProperties._prior attribute! It must "
                f"be of type PriorBase or None, but found {type(self._prior)}.")

        if not (type(self._value) in [float, int] or self._value is None):
            raise TypeError(
                f"Found invalid ParameterProperties._value attribute! It must "
                f"be of type float/int or None, but found {type(self._value)}.")

        # -------------------------------- #
        #           Cross checks           #
        # -------------------------------- #

        if self._index is not None:
            # in this case, we have a latent parameter
            if self._value is not None:
                raise RuntimeError(
                    f"ParameterProperties._index and ParameterProperties._value"
                    f" are both given (_index={self._index} and _value="
                    f"{self._value}), but one of them must be None!")
            if self._prior is None:
                raise RuntimeError(
                    f"ParameterProperties._index and ParameterProperties._prior"
                    f" are both given (_index={self._index} and prior="
                    f"{self._prior}), but one of them must be None!")

        else:
            # in this case, we have a constant parameter
            if self._value is None:
                raise RuntimeError(
                    f"ParameterProperties._index and ParameterProperties._value"
                    f" are both None, but one of them must be not None!")
            if self._prior is not None:
                raise RuntimeError(
                    f"ParameterProperties._index is None while Parameter"
                    f"Properties._prior is given ({self._prior}). This "
                    f"combination is not valid. Either the index must also be "
                    f"given, or the prior must also be None.")

    @property
    def index(self):
        """Access self._index from outside via self.index."""
        return self._index

    @index.setter
    def index(self, value):
        """Raise a specific error when trying to directly set self.index."""
        raise AttributeError(
            "Changing a parameter's index directly is prohibited!")

    @property
    def type(self):
        """Access self._type from outside via self.type."""
        return self._type

    @type.setter
    def type(self, value):
        """Raise a specific error when trying to directly set self.type."""
        raise AttributeError(
            "Changing a parameter's type directly is prohibited!")

    @property
    def role(self):
        """Adds a pseudo-attribute self.role, which allows a convenient check
           on whether a parameter is latent or not."""
        return 'latent' if self._index is not None else 'const'

    @role.setter
    def role(self, value):
        """Raise a specific error when trying to directly set self.role."""
        raise AttributeError(
            "You cannot change a parameter's role directly! Use "
            "InferenceProblem.change_parameter_role instead.")

    @property
    def is_latent(self):
        """Adds a pseudo-attribute self.is_latent, which allows a convenient
           check on whether a parameter is latent or not."""
        return self._index is not None

    @property
    def is_const(self):
        """Adds a pseudo-attribute self.is_const, which allows a convenient
           check on whether a parameter is constant or not."""
        return not self.is_latent

    @property
    def prior(self):
        """Access self._prior from outside via self.prior."""
        return self._prior

    @prior.setter
    def prior(self, value):
        """Raise a specific error when trying to directly set self.prior."""
        raise AttributeError(
            "Changing a parameter's prior directly is prohibited!")

    @property
    def value(self):
        """Access self._value from outside via self.value."""
        return self._value

    @value.setter
    def value(self, value):
        """Raise a specific error when trying to directly set self.value."""
        raise AttributeError(
            "Changing a parameter's value directly is prohibited!")



# third party imports
import numpy as np
import torch as th
from pyro.infer import NUTS, MCMC
import arviz as az
from loguru import logger
import time
import contextlib

# local imports
from probeye.inference.torch_.priors import translate_prior_template
from probeye.inference.torch_.noise_models import translate_noise_model
from probeye.subroutines import len_or_one
from probeye.subroutines import pretty_time_delta, stream_to_logger
from probeye.subroutines import print_dict_in_rows


class PyroSolver:
    """Solver routines based on pyro/torch for an InferenceProblem."""

    def __init__(self, problem, seed=1, show_progress=True):
        """
        Parameters
        ----------
        problem : obj[InferenceProblem]
            Describes the inference problem including e.g. parameters and data.
        seed : int, optional
            Random state used for random number generation.
        show_progress : bool, optional
            When True, the progress of a solver routine will be shown (for
            example as a progress-bar) if such a feature is available.
            Otherwise, the progress will not shown.
        """

        # attributes from arguments
        self.show_progress = show_progress
        self.seed = seed

        # the following attribute will be set after the solver was run
        self.raw_results = None

        # the problem is copied, and in the copy, the experimental data is
        # reformatted from numpy-arrays to torch-tensors
        self.problem = problem.transform_experimental_data(f=th.from_numpy)

        # each noise model must be connected to the relevant experiment_names
        self.problem.assign_experiments_to_noise_models()

        # the dictionary dependency_dict will contain all latent parameter names
        # as keys; the value of each key will be a list with latent hyper-
        # parameters of the latent parameter's prior
        logger.debug("Checking parameter's dependencies")
        dependency_dict = dict()
        for prm_name in self.problem.parameters.latent_prms:
            dependency_dict[prm_name] = []
            hyperparameters =\
                self.problem.parameters[prm_name].prior.hyperparameters
            for prior_prm_name in hyperparameters:
                if prior_prm_name in self.problem.parameters.latent_prms:
                    dependency_dict[prm_name].append(prior_prm_name)

        # this makes sure that the items in dependency are in an order that they
        # ca be sampled from beginning (index 0) sequentially until the last
        # item without encountering any dependency problems (i.e. that another
        # parameter has to be sampled before the considered prm can be sampled)
        consistent = False
        while not consistent:
            consistent = True
            idx_latent_dependencies =\
                [i for i, v in enumerate(dependency_dict.values())
                 if len(v) > 0]
            for idx in idx_latent_dependencies:
                key_idx = [*dependency_dict.keys()][idx]
                for dependency in dependency_dict[key_idx]:
                    if key_idx in dependency_dict[dependency]:
                        raise RuntimeError(
                            f"Found circular dependency between {key_idx} and "
                            f"{dependency}!")
                    idx_dependency = [*dependency_dict.keys()].index(dependency)
                    if idx_dependency > idx:
                        consistent = False
                        tuples = [*dependency_dict.items()]
                        tuples[idx], tuples[idx_dependency] = \
                            tuples[idx_dependency], tuples[idx]
                        dependency_dict = dict(tuples)

        # translate the prior definitions to objects with computing capabilities
        logger.debug("Translating problem's priors")
        self.priors = {}
        for prm_name in dependency_dict.keys():
            prior_template = self.problem.parameters[prm_name].prior
            self.priors[prior_template.ref_prm] = \
                translate_prior_template(prior_template)

        # translate the general noise model objects into solver specific ones
        logger.debug("Translating problem's noise models")
        self.noise_models = []
        for noise_model_base in self.problem.noise_models:
            self.noise_models.append(translate_noise_model(noise_model_base))

        # translate the problem's forward models into torch compatible ones
        logger.debug("Wrapping problem's forward models")
        for fwd_model_name in self.problem.forward_models.keys():
            setattr(self.problem.forward_models[fwd_model_name], 'call',
                    self._translate_forward_model(
                        self.problem.forward_models[fwd_model_name]))

    @staticmethod
    def _only_values(func):
        """
        This function wrapper is required for the Autograd.apply function which
        is returned by translate_forward_model. While the forward model's
        response method takes a dict as input, the Autograd.apply function
        which will essentially take the forward model's response method's place
        requires the numeric values of inp given as positional arguments. This
        pre-processing step is done by this function wrapper.
        """
        def wrapper(inp):
            return func(*inp.values())
        return wrapper

    def _translate_forward_model(self, forward_model):
        """
        Translates a given forward model (based on non-tensor in/outputs) to a
        torch-compatible forward model based on tensors.

        Parameters
        ----------
        forward_model : ForwardModelBase
            The forward model object. Note that this object needs to have a
            jacobian method to make it translatable.
        """

        class Autograd(th.autograd.Function):

            @staticmethod
            def forward(ctx, *values):
                """
                Modified from the th.autograd.Function docstring: This function
                must accept a context ctx as the first argument, followed by any
                number of arguments (tensors or other types). The context can be
                used to store arbitrary data that can be then retrieved during
                the backward pass.

                Parameters
                ----------
                ctx : torch.autograd.function.AutogradBackward
                    This context-object is for derivative-bookkeeping.
                values : iterable[th.tensor]
                    The values of the inp-dict of the inp-argument of the
                    forward model's 'response' method in the same order.
                """

                # the idea is that the given forward model only takes numeric
                # (non-tensor) inputs; so there needs to be a conversion which
                # takes place here; also, the given values must be rearranged
                # in the dict-format required by forward_model's response method
                inp = {}
                keys = [isens.name for isens in forward_model.input_sensors] +\
                       [*forward_model.prms_def.values()]
                for key, value in zip(keys, values):
                    if th.is_tensor(value):
                        inp[key] = value.detach().numpy()
                    else:
                        inp[key] = value

                # evaluate the forward model and its jacobian for the given
                # input parameters; this is where the forward model is evaluated
                # in its original setup without any tensors
                response_dict = forward_model.response(inp)
                jac_dict = forward_model.jacobian(inp)
                jac_numpy = forward_model.jacobian_dict_to_array(inp, jac_dict)

                # now we have the forward model's response in dict format;
                # however, in this format it cannot be processed here, so we
                # will put all its numeric results in one long vector
                n1 = len(forward_model.output_sensors)
                n1_times_n3, _ = jac_numpy.shape
                n3 = n1_times_n3 // n1
                response_numpy = np.zeros(n1_times_n3)
                response_structure = dict()
                for i, (key, value) in enumerate(response_dict.items()):
                    n_elements = len_or_one(value)
                    idx_start = i * n3
                    idx_end = idx_start + n_elements
                    response_numpy[idx_start: idx_end] = value
                    response_structure[key] = n_elements
                forward_model.response_structure = response_structure

                # now, we need to translate back to the tensor-world
                jac_torch = th.from_numpy(jac_numpy)
                y_torch = th.from_numpy(response_numpy)

                # the jacobian is recorded, and the model evaluation returned
                ctx.save_for_backward(jac_torch)
                return y_torch

            @staticmethod
            def backward(ctx, *dl_dy_tuple):
                """
                From the torch.autograd.Function docstring (modified): It must
                accept a context object as the first argument (ctx), followed by
                as many outputs as the 'forward' method of this class returned
                (None will be passed for non-tensor outputs of the 'forward'
                method), and it should return as many tensors, as there were
                inputs to 'forward' method. Each argument of this method (i.e.,
                the 'backward' method) is the gradient w.r.t the given output,
                and each returned value should be the gradient w.r.t. the
                corresponding input. If an input is either not a Tensor or a
                Tensor not requiring gradients, you can just pass None as a
                gradient for that particular input.

                Parameters
                ----------
                ctx : torch.autograd.function.AutogradBackward
                    This context-object is for derivative-bookkeeping.
                dl_dy_tuple : tuple(torch.Tensor)
                    The jacobian of the likelihood with respect to the forward
                    model evaluations, i.e., dl/dy. Since the likelihood returns
                    a scalar value, this is a tuple of length 1.

                Returns
                -------
                tuple
                    Each element i of the tuple is a tensor that represents the
                    derivative dl/dtheta_i. The order must correspond to the
                    order in the input arguments of the forward-method after the
                    ctx argument. An element is None, if it does not influence
                    the value of the likelihood function.
                """

                # this is dy/dtheta (the gradient with respect to the forward
                # model evaluation) stored from the forward pass; note that the
                # '[0]' at the end is necessary since ctx.saved_tensors is a
                # tuple of all saved tensors; here, there is only one element
                # in it, but we still need to request it
                dy_dtheta = ctx.saved_tensors[0]

                # this is dl/dtheta (the gradient of the likelihood function
                # with respect to the model parameters theta); note that
                # dl_dy_tuple is a tuple of length 1 since the likelihood is a
                # scalar (and not a vector-valued) function
                dl_dy = dl_dy_tuple[0]

                # now dl/dtheta (grad) is computed using the chain rule (which
                # reads dl/dy * dy/dtheta here)
                grad_total = th.matmul(dl_dy, dy_dtheta)

                # the returned gradients must be in tuple form; note that
                # ctx.needs_input_grad is a list with True/False elements of
                # the same lengths as there are parameters in theta; elements
                # in theta that are not being sampled (for example the value of
                # an input sensor of the forward model) are not required to have
                # their gradients evaluated, so these elements will have 'False'
                # entries in ctx.needs_input_grad
                return_val = tuple(
                    None if not required else grad for required, grad
                    in zip(ctx.needs_input_grad, th.flatten(grad_total)))
                return return_val

        return self._only_values(Autograd.apply)

    def get_theta_samples(self):
        """
        Provides a list of latent-parameter samples in form of torch.Tensors.

        Returns
        -------
        list[torch.Tensor]
            The sampled values based on the latent parameter's priors.
        """
        # even if a list is returned by this function, we initialize a dict here
        # to enable the action in the if-statement in the loop below
        pyro_parameter_samples = dict()
        # we have to sample each parameter for which a prior is defined
        for ref_prm, prior_obj in self.priors.items():
            hyperprms_dict = {}
            for name in prior_obj.hyperparameters.keys():
                if name in pyro_parameter_samples:
                    hyperprms_dict[name] = pyro_parameter_samples[name]
                else:
                    # this is the default case, where the prior's parameters
                    # (i.e. the hyperparameters) are simply constants
                    hyperprms_dict[name] = self.problem.parameters[name].value
            # this is where the parameter's sample is generated with pyro
            pyro_parameter_samples[ref_prm] = prior_obj.sample(hyperprms_dict)
        return [*pyro_parameter_samples.values()]

    def loglike(self, theta):
        """
        Evaluates the log-likelihood function of the problem at theta.

        Parameters
        ----------
        theta : list[torch.Tensor]
            A vector of pyro.samples (i.e. tensors) for which the log-likelihood
            function should be evaluated.

        Returns
        -------
        ll : torch.Tensor
            The evaluated log-likelihood function for the given theta-vector.
        """
        # compute the contribution to the log-likelihood function for each noise
        # model and sum it all up
        for noise_model in self.noise_models:
            # compute the model response for the noise model's experiment_names
            model_response = self.problem.evaluate_model_response(
                theta, noise_model.experiment_names)
            # get the tensors for the noise model's parameters
            prms_noise = self.problem.get_parameters(
                theta, noise_model.prms_def)
            # evaluate the loglike-contribution for the noise model
            noise_model.sample_cond_likelihood(model_response, prms_noise)

    def posterior_model(self):
        """
        Returns the sampled log-likelihood in form of a torch.Tensor.

        Returns
        -------
        list[torch.Tensor]
            The sampled values based on the latent parameter's priors.
        """
        theta = self.get_theta_samples()
        return self.loglike(theta)

    def run_mcmc(self, n_walkers=1, n_steps=300, n_initial_steps=30,
                   step_size=0.1, **kwargs):
        """
        Runs MCMC with NUTS kernel for the InferenceProblem the PyroSolver was
        initialized with and returns the results as an arviz InferenceData obj.

        Parameters
        ----------
        n_walkers : int, optional
            Number of walkers used by the estimator.
        step_size : float, optional
            The step size of the NUTS sampler.
        n_steps: int, optional
            The number of steps the sampler takes.
        n_initial_steps: int, optional
            The number of steps for the burn-in phase.
        kwargs : dict
            Additional keyword arguments passed to NUTS.

        Returns
        -------
        inference_data : obj[arviz.data.inference_data.InferenceData]
            Contains the results of the sampling procedure.
        """

        # log which solver is used
        logger.info(
            f"Solving problem using pyro's NUTS sampler with {n_initial_steps} "
            f"+ {n_steps} samples, ...")
        logger.info(f"... {n_walkers} chains and a step size of {step_size:.3f}")
        if kwargs:
            logger.info("Additional NUTS options:")
            print_dict_in_rows(kwargs, printer=logger.info)
        else:
            logger.info("No additional NUTS options specified")

        # prepare the sampling with the requested parameters
        logger.debug("Setting up NUTS sampler")
        th.manual_seed(self.seed)
        kernel = NUTS(self.posterior_model, step_size=step_size, **kwargs)
        logger.debug("Starting sampling (warmup + main)")
        time.sleep(0.1)  # for logging; otherwise no time for new line

        # this is where the actual sampling happens
        start = time.time()
        mcmc = MCMC(kernel,
                    num_samples=n_steps,
                    warmup_steps=n_initial_steps,
                    num_chains=n_walkers,
                    disable_progbar=not self.show_progress)
        mcmc.run()
        end = time.time()

        # log out the results of the process
        runtime_str = pretty_time_delta(end - start)
        logger.info(f"Sampling of the posterior distribution completed: "
                    f"{n_steps} steps and {n_walkers} chains.")
        logger.info(f"Total run-time (including warmup): {runtime_str}.")
        logger.info("")
        logger.info("Summary of sampling results")
        with contextlib.redirect_stdout(stream_to_logger('INFO')):
            mcmc.summary()
        self.raw_results = mcmc

        # translate the results to a common data structure and return it
        var_names = self.problem.get_theta_names(tex=False)
        var_names_tex = self.problem.get_theta_names(tex=True)
        name_dict = {var_name: var_name_tex for var_name, var_name_tex
                     in zip(var_names, var_names_tex)}
        inference_data = az.from_pyro(mcmc)
        inference_data.rename(name_dict, groups='posterior', inplace=True)

        return inference_data



# third party imports
import numpy as np

# local imports
from probeye.subroutines import translate_prms_def


class PriorBase:
    """
    Template class for prior definitions. Note that the main motivation of how
    this class is implemented was to avoid storing any numeric values for any of
    the priors parameters within the prior object.
    """

    def __init__(self, ref_prm, prms_def, name, prior_type):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings, or list of one-element-dicts defining the prior's
            parameter names. E.g. ['loc_a', 'scale_a'] or {'loc_a': 'loc_a',
            'scale_a': 'scale_a'}. The latter example is the notation for the
            use of global and local names, which should not be necessary for
            the definition of prior-parameters. A special case is the
            uninformative prior (see below) which hasn't got an parameters. So,
            in this case prms_def might also be None.
        name : string
            Defining the priors name.
        prior_type : string
            Stating the prior's type, e.g. "normal distribution". This is just
            used for printing information on the prior.
        """
        # write arguments to attributes
        self.ref_prm = ref_prm
        self.name = name
        self.prior_type = prior_type

        # these dictionaries contain the <global name>:<local name> pairs of the
        # prior's parameters with and without the reference variable; if for
        # example you define a normal prior for parameter 'a' with location,
        # then 'a' is the reference variable; note that the conversion to a
        # dictionary via list2dict is due to the possibility of using local
        # parameter names, which however is not intended to be used for priors
        self.hyperparameters, _ = translate_prms_def(prms_def)
        self.prms_def, _ = translate_prms_def(
            [ref_prm] + [*self.hyperparameters.values()])

    def __str__(self):
        """
        Allows printing an object of this class.

        Returns
        -------
        s : string
            A string containing details on the respective prior.
        """
        s = f"{self.prior_type} for '{self.ref_prm}', prms={self.prms_def}"
        return s

    def plot(self, ax, prms,  x=None, n_points=200, n_sigma=2,
             color='darkorange'):
        """
        Plots the prior-pdf to a given axis object.

        Parameters
        ----------
        ax : obj[matplotlib.axes]
            The axis object to plot the prior-pdf on.
        prms : obj[Parameters]
            The parameters of the problem at hand.
        x : numpy.ndarray, None, optional
            The points where the prior-pdf should be evaluated at. If None is
            given, x will be derived from the x-limits of the given ax-object.
        n_points : int, optional
            The number of points of the prior-pdf graph. Only effective when
            x is None.
        n_sigma : float, int, optional
            Defines the x-range of a normal prior via mean plus/minus n_sigma
            times the standard deviation.
        color : str, optional
            The line-color of the prior-pdf's graph.
        """

        if self.prior_type == 'normal':
            mu = prms[self.prms_def[f"loc_{self.ref_prm}"]].value
            sigma = prms[self.prms_def[f"scale_{self.ref_prm}"]].value
            # proceed, only if both values are constants and not latent
            # parameters themselves
            if mu and sigma:
                if x is None:
                    x = np.linspace(
                        mu - n_sigma * sigma, mu + n_sigma * sigma, n_points)
                y = 1 / (sigma * np.sqrt(2 * np.pi))
                y *= np.exp(-0.5 * ((x - mu) / sigma) ** 2)
                ax.plot(x, y, label='prior', color=color)

        elif self.prior_type == 'uniform':
            a = prms[self.prms_def[f"low_{self.ref_prm}"]].value
            b = prms[self.prms_def[f"high_{self.ref_prm}"]].value
            # proceed, only if both values are constants and not latent
            # parameters themselves
            if a and b:
                y = np.zeros(n_points)
                y[1:-1] = np.ones(n_points - 2) / (b - a)
                if x is None:
                    x = np.linspace(a, b, n_points)
                    y[0], y[-1] = 0, 0
                else:
                    y[0] = 0 if (x[0] <= a) else y[1]
                    y[-1] = 0 if (x[-1] >= b) else y[-2]
                ax.plot(x, y, label='prior', color=color)



# third party imports
import numpy as np

# local imports
from probeye.subroutines import make_list, translate_prms_def


class NoiseModelBase:

    def __init__(self, dist, prms_def, sensors, name=None, corr=None,
                 corr_model=None, noise_type='additive'):
        """
        Parameters
        ----------
        dist : str
            A string specifying the probability distribution the noise model
            is based on, e.g. 'normal'.
        prms_def : str, list[str], dict
            A list of parameter names (strings) defining how a noise parameter
            vector given to the loglike_contribution method is interpreted. E.g.
            prms_def = ['mu', 'sigma'] means that the noise parameter vector
            has two elements, the first of which gives the value of 'mu' and the
            second gives the value of 'sigma'.
        sensors : obj[Sensor], list[Sensor]
            Sensor objects that are required to evaluate the noise model.
        name : str, None, optional
            Unique name of the noise model. This name is None, if the user does
            not specify it when adding the noise model to the problem. It is
            then named automatically before starting the inference engine.
        corr : dict, None, optional
            Defines the correlation model. So far this is just a placeholder.
            It is not clear yet how exactly the correlation should be defined.
            When it is set to None, all sensors/sensor elements are independent.
        corr_model : str, optional
            Defines the correlation function to be used in case corr isn't None.
        noise_type : str, optional
            Either 'additive', 'multiplicative' or 'other'. Defines if the error
            is computed via [prediction - measurement] ('additive') or via
            [prediction/measurement-1] ('multiplicative') or in some 'other'
            i.e., non-standard fashion.
        """
        self.dist = dist
        self.prms_def, self.prms_dim = translate_prms_def(prms_def)
        self.sensors = make_list(sensors)
        self.sensor_names = [sensor.name for sensor in self.sensors]
        self.name = name
        self.corr = corr
        self.corr_model = corr_model
        self.noise_type = noise_type
        # this is a list of experiment names, that relate to the noise model;
        # the list will be filled after experiments have been added to the
        # InferenceProblem and the problem definition is complete; in this case
        # call InferenceProblem.assign_experiments_to_noise_models() to fill
        # this list with the corresponding names
        self.experiment_names = []
        # as soon as defined, this attribute will be a pointer to the inference
        # problems experiments (it will be used for consistency checks)
        self.problem_experiments = {}

        # set the error_function depending on the noise-type
        if noise_type == "additive":
            self.error_function = self.error_function_additive
        elif noise_type == "multiplicative":
            self.error_function = self.error_function_multiplicative
        elif noise_type == "other":
            self.error_function = self.error_function_other
        else:
            raise ValueError(
                f"Encountered unknown noise_type: '{noise_type}'. The noise"
                f"_type must be either 'additive', 'multiplicative' or 'other'."
            )

    def add_experiments(self, experiment_names_):
        """
        Adds experiment names to the noise model. When the noise model is
        evaluated it will only be evaluated for those experiments added here.

        Parameters
        ----------
        experiment_names_ : str, list[str]
            Names (strings) of experiments from the InferenceProblem that should
            be added to the noise model.
        """
        # check if the given experiments are compatible with the noise model
        # with respect to the sensors
        experiment_names = make_list(experiment_names_)
        forward_models = set()
        for exp_name in experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            forward_models.add(exp_dict['forward_model'])
            sensor_names_exp = [*exp_dict['sensor_values'].keys()]
            for sensor_name in self.sensor_names:
                if sensor_name not in sensor_names_exp:
                    raise RuntimeError(
                        f"Experiment '{exp_name}' does not contain a sensor "
                        f"'{sensor_name}' which is required for the evaluation "
                        f"of the noise model.")
        # check if the given experiments all refer to one forward model
        if len(forward_models) > 1:
            raise RuntimeError(
                f"The given experiments refer to more than one forward model!")
        # check if one of the given experiments have been added before
        for exp_name in experiment_names:
            if exp_name in self.experiment_names:
                raise RuntimeError(
                    f"The experiment '{exp_name}' has already been added to "
                    f"this noise model. Something might be wrong here.")
        self.experiment_names += experiment_names

    def error(self, model_response_dict):
        """
        Computes the model error for all of the noise model's experiments and
        returns them in a dictionary that is sorted by output sensor_values.

        Parameters
        ----------
        model_response_dict : dict
            The first key is the name of the experiment. The values are dicts
            which contain the forward model's output sensor's names as keys
            have the corresponding model responses as values.

        Returns
        -------
        model_error : dict
            A dictionary with the keys being the noise model's sensor names, and
            1D numpy arrays representing the model errors as values.
        """
        # prepare the dictionary keys
        model_error_dict = {name: np.array([]) for name in self.sensor_names}

        # fill the dictionary with model error vectors
        for exp_name in self.experiment_names:
            exp_dict = self.problem_experiments[exp_name]
            ym_dict = model_response_dict[exp_name]
            ye_dict = exp_dict['sensor_values']
            me_dict = self.error_function(ym_dict, ye_dict)
            model_error_dict =\
                {name: np.append(model_error_dict[name], me_dict[name])
                 for name in self.sensor_names}

        return model_error_dict

    def error_function_additive(self, ym_dict, ye_dict):
        """
        Evaluates the additive model error for each of the noise model' sensors.

        Parameters
        ----------
        ym_dict : dict
            The computed values for the model's output sensor_values.
        ye_dict : dict
            The measured values for the model's output sensor_values.

        Returns
        -------
        error_dict : dict
            The computed model error for the model's output sensor_values.
        """
        # for each sensor, its own error metric is used to compute the error
        error_dict = {name: ym_dict[name] - ye_dict[name]
                      for name in self.sensor_names}
        return error_dict

    def error_function_multiplicative(self, ym_dict, ye_dict):
        """
        Evaluates the multiplicative model error for each of the noise model's
        sensors.

        Parameters
        ----------
        ym_dict : dict
            The computed values for the model's output sensor_values.
        ye_dict : dict
            The measured values for the model's output sensor_values.

        Returns
        -------
        error_dict : dict
            The computed model error for the model's output sensor_values.
        """
        # for each sensor, its own error metric is used to compute the error
        error_dict = {name: ym_dict[name] / ye_dict[name] - 1.0
                      for name in self.sensor_names}
        return error_dict

    def error_function_other(self, ym_dict, ye_dict):
        """
        Non-standard error function self.error_function will point to when
        self.noise_type is set to 'other'. See self.error_function for more
        information.
        """
        raise NotImplementedError(
            "Your model does not have an non-standard error_function-method "
            "yet. If you want to use it, you need to implement it first.")

    def loglike_contribution(self, model_response_dict, prms):
        """
        Evaluates the log-likelihood function for the given model error and
        the given noise parameter vector. This method has to be overwritten.

        Parameters
        ----------
        model_response_dict : dict
            The first key is the name of the experiment. The values are dicts
            which contain the forward model's output sensor's names as keys
            have the corresponding model responses as values.
        prms : ParameterList-object
            Dictionary-like object containing parameter name:value pairs.

        Returns
        -------
        ll : float
            The evaluated log-likelihood function.
        """
        raise NotImplementedError(
            "Your model does not have a loglike_contribution-method. You " +
            "need to define this method so you can evaluate your noise model.")


class NormalNoiseModel(NoiseModelBase):
    """
    A general Gaussian (normal) noise model with or without correlations.
    """

    def __init__(self, prms_def, sensors, name=None, corr=None, corr_model=None,
                 noise_type='additive'):
        """
        See docstring of NoiseModelBase for information on the input arguments.
        """

        # initialize the base class with given input
        super().__init__('normal', prms_def, sensors, name=name, corr=corr,
                         corr_model=corr_model, noise_type=noise_type)

        # check that at the standard deviation is provided (this can be either
        # as a constant or a latent parameter, but it has to be given)
        if 'std' not in [*self.prms_def.values()]:
            raise RuntimeError(
                "The standard deviation 'std' was not provided in prms_def!")

        # the mean value(s) do not have to be stated explicitly; if they are not
        # given, the are assumed to be zero
        self.zero_mean = True
        if 'mean' in [*self.prms_def.values()]:
            self.zero_mean = False



# standard library
import copy as cp
import numpy as np

# local imports
from probeye.subroutines import translate_prms_def
from probeye.subroutines import make_list
from probeye.subroutines import len_or_one


class ForwardModelBase:
    """
    This class serves as a base class for any forward model. When you want to
    define a specific forward model, you need to derive your own class from this
    one, and then define the '__call__' method. The latter essentially describes
    the model function mapping the model input to the output.
    """

    def __init__(self, prms_def_, input_sensors, output_sensors):
        """
        Parameters
        ----------
        prms_def_ : str, list, dict
            Contains the model's latent parameter names. The list may only
            contain strings or one-element dictionaries. It could look, for
            example, like [{'a': 'm'}, 'b']. The one-element dictionaries
            account for the possibility to define a local name for a latent
            parameter that is different from the global name. In the example
            above, the latent parameter with the global name 'a' will be
            referred to as 'm' within the model. So, the one-element dicts have
            the meaning {<global name>: <local name>}. String-elements are
            interpreted as having similar local and global names. Note that the
            local-name option will not be required most of the times. The input
            from global to local name can also be provided as a dict. In the
            example above it would look like {'a': 'm', 'b': 'b'}.
        input_sensors : Sensor, list[Sensor]
            Contains sensor-objects structuring the model input.
        output_sensors : Sensor, list[Sensor]
            Contains sensor-objects structuring the model output.
        """

        # convert the given parameter names to a dictionary with global names
        # as keys and local names as values
        self.prms_def, self.prms_dim = translate_prms_def(prms_def_)

        # other attributes
        self.input_sensors = make_list(input_sensors)
        self.output_sensors = make_list(output_sensors)

        # this attributes might be used by inference engines that need a forward
        # model wrapper, which only returns numeric vectors; for reconstructing
        # the response dictionary from the numeric vector, one needs to know the
        # response dictionary's structure; this dictionaries will then contain
        # the same keys as the response method's return dictionary, while the
        # values will be the number of elements contained in the values; for
        # example {'x': np.array([0, 0.1, 0.2]), 'a': 3.7} will have a structure
        # of {'x': 3, 'a': 1}; this attr. is not used by all inference engines
        self.response_structure = dict()

    @property
    def input_sensor_names(self):
        """Provides input_sensor_names attribute."""
        return [sensor.name for sensor in self.input_sensors]

    @property
    def output_sensor_names(self):
        """Provides input_sensor_names attribute."""
        return [sensor.name for sensor in self.output_sensors]

    def response(self, inp):
        """
        Evaluates the model response and provides computed results for all of
        the model's output sensors. This method must be overwritten by the user.

        Parameters
        ----------
        inp : dict
            Contains both the exp. input data and the  model's parameters. The
            keys are the names, and the values are their numeric values.

        Returns
        -------
        dict
            Contains the model response (value) for each output sensor,
            referenced by the output sensor's name (key).
        """
        raise NotImplementedError(
            "Your model does not have a proper 'response'-method yet. You need "
            "to define this method, so you can evaluate your model.")

    def __call__(self, inp):
        """
        Evaluates the forward model either via calling the original or the
        wrapped response method, depending on the current definition of the
        call-method. This method is equivalent to calling self.response as long
        as the call-method was not overwritten. If it was overwritten, however,
        this method (the __call__-method) makes sure, that the returned value
        has the same format as the returned value by the response-method. See
        self.response for a docstring on its parameters/return values.
        """
        res_ori = self.call(inp)
        if type(res_ori) is dict:
            # in this case it is assumed that res_ori is the dictionary returned
            # by the response-method
            res = res_ori
        else:
            # in this case, the returned value is assumed to be a numeric vector
            # which needs to be translated back to the dictionary structure
            res = self.response_structure
            i = 0
            for key in self.response_structure.keys():
                n_numbers = self.response_structure[key]
                res[key] = res_ori[i:i + n_numbers]
                i += n_numbers
        return res

    def call(self, inp):
        """
        This function can be used by inference engines to wrap the forward
        model's response. This can be done by overwriting this method by one
        that might do something before and after calling self.response(inp).
        See self.response for a docstring on its parameters/return values.
        """
        return self.response(inp)

    def jacobian(self, inp):
        """
        Numerically computes the Jacobian matrix of the forward model and
        returns it in form of a dictionary. Note that this method should be
        overwritten, if there is a more efficient way to compute the jacobian,
        for example, when one can compute the Jacobian analytically.

        Parameters
        ----------
        inp : dict
            Contains both the exp. input data and the  model's parameters. The
            keys are the names, and the values are their numeric values.

        Returns
        -------
        jac_dict : dict or numpy.ndarray
            The Jacobian matrix in dict-form: The keys are the names of the
            forward model's output sensors. The values are dictionaries with the
            forward model's input channel's names as keys and the derivatives as
            values. For example, the element jac['y']['a'] would give the
            derivative dy/da, and jac['y'] would give the gradient of the
            forward model's y-computation with respect to theta (i.e., the
            input channels) in a dictionary-format.
        """
        # prepare the dictionary; this structure needs to be external from the
        # main loop below since the filling of the dictionary could only be
        # efficiently done in the format jac_dict[prm_name][os_name] which is
        # less readable; the format created in the implemented way is easier to
        # to read since jac['y']['a'] corresponds to dy/da in jac['y'] is the
        # gradient of y with respect to theta
        jac_dict = {}
        for output_sensor in self.output_sensors:
            jac_dict[output_sensor.name] = {}
            for prm_name in inp.keys():
                jac_dict[output_sensor.name][prm_name] = None

        # eps is the machine precision; it is needed to compute the step size of
        # the central difference scheme below; note that this refers to single
        # precision (float32) since the processed arrays might be defined in
        # float32, in which case using the eps of double precision (float64)
        # would not work since the step size would be too small
        eps = np.finfo(np.float32).eps
        # the following evaluations are needed in the for-loop; they are put
        # here so they are not repeatedly evaluated (to the same value) during
        # the for-loop
        sqrt_eps = np.sqrt(eps)
        response_dict_center = self.response(inp)
        inp_right = cp.copy(inp)
        for prm_name, prm_value in inp.items():
            x = inp[prm_name]
            # the following formula for the step size is NOT taken from the
            # literature; in the literature, a common recommended choice for the
            # step size h given x is not 0 is h = sqrt_eps * x, see for example
            # https://en.wikipedia.org/wiki/Numerical_differentiation; we added
            # the term '+ sqrt_eps' below to also cover the cases where x
            # actually is zero (or very close to 0)
            h = sqrt_eps * x + sqrt_eps
            inp_right[prm_name] = x + h
            response_dict_right = self.response(inp_right)
            for output_sensor in self.output_sensors:
                # the simple forward scheme should be sufficient for most
                # applications since the Jacobian will only be used as info for
                # choosing the next sample; for that purpose it is secondary if
                # it contains small numerical errors
                jac_dict[output_sensor.name][prm_name] =\
                    (response_dict_right[output_sensor.name] -
                     response_dict_center[output_sensor.name]) / h
            inp_right[prm_name] = inp[prm_name]  # resetting perturbed value
        return jac_dict

    def jacobian_dict_to_array(self, inp, jac_dict):
        """
        Converts the Jacobian in dict-format (computed by the above 'jacobian'
        method) into a numpy array. This method is external to the above
        'jacobian' method, so that it is easier for a user to overwrite the it
        (i.e., the 'jacobian' method) without also having to define the
        conversion into an array.

        Parameters
        ----------
        inp : dict
            See docstring of the 'jacobian'-method above.
        jac_dict : dict
            See docstring of the 'jacobian'-method above.

        Returns
        -------
        jac : numpy.ndarray
            Similar structure as the conventional Jacobi matrix with respect to
            the columns and rows (i.e. the rows are the different gradients and
            the columns are derivatives with respect to one fixed parameter).
        """

        # n1 is the number of the forward model's output sensors; n2 is the
        # number of the forward model's input channels, i.e., the number of
        # input sensors and the number of the forward model's parameters;
        # finally, n3 is the maximum number of elements in the n2 input
        # channels; the model's parameters are usually scalars, but the
        # input sensors might be vectors with more than one element
        n1 = len(self.output_sensors)
        n2 = len(inp)
        n3 = max([len_or_one(v) for v in [*inp.values()]])
        jac = np.zeros((n1 * n3, n2))
        for i, prm_dict in enumerate([*jac_dict.values()]):
            idx_start = i * n3
            for j, derivative in enumerate([*prm_dict.values()]):
                idx_end = idx_start + len_or_one(derivative)
                jac[idx_start: idx_end, j] = derivative

        return jac



# third party imports
import arviz as az
import numpy as np
import matplotlib.pyplot as plt


def create_pair_plot(inference_data, problem, plot_with="arviz",
                     plot_priors=True, focus_on_posterior=False, kind="kde",
                     figsize=(9, 9), textsize=10, true_values=None,
                     show_legends=True, **kwargs):
    """
    Creates a pair-plot for the given inference data using arviz.

    Parameters
    ----------
    inference_data : obj[arviz.data.inference_data.InferenceData]
        Contains the results of the sampling procedure.
    problem : obj[InferenceProblem]
        The inference problem the inference data refers to.
    plot_with : {'arviz', 'seaborn', 'matplotlib'}, optional
        Defines the python package the plot will be generated with.
    plot_priors : bool, optional
        If True, the prior-distributions are included in the marginal subplots.
        Otherwise the priors are not shown.
    focus_on_posterior : bool, optional
        If True, the marginal plots will focus on the posteriors, i.e., the
        range of the horizontal axis will adapt to the posterior. This might
        result in just seeing a fraction of the prior distribution (if they are
        included). If False, the marginal plots will focus on the priors, which
        will have a broader x-range. If plot_priors=False, this argument has no
        effect on the generated plot.
    kind : str, optional
        Type of plot to display ('scatter', 'kde' and/or 'hexbin').
    figsize : tuple, None, optional
        Defines the size of the generated plot in the default unit. If None is
        chose, the figsize will be derived automatically.
    textsize : float, optional
        Defines the font size in the default unit.
    true_values : None, dict, optional
        Used for plotting 'true' parameter values. Keys are the parameter names
        and values are the values that are supposed to be shown in the marginal
        plots.
    show_legends : bool, optional
        If True, legends are shown in the marginal plots. Otherwise no legends
        are included in the plot.
    kwargs
        Additional keyword arguments passed to arviz' pairplot function.
    """

    if plot_with == 'arviz':

        # these names will appear on the axis labels
        var_names = problem.get_theta_names(tex=True)

        # set default value for kde_kwargs if not given in kwargs; note that
        # this default value is mutable, so it should not be given as a default
        # argument in create_pair_plot
        if 'kde_kwargs' not in kwargs:
            kwargs['kde_kwargs'] = {'contourf_kwargs': {"alpha": 0},
                                    'contour_kwargs': {"colors": None}}

        # process true_values if specified
        if true_values is not None:
            reference_values = dict()
            for prm_name, value in true_values.items():
                key = problem.parameters[prm_name].tex
                reference_values[key] = value
            kwargs['reference_values'] = reference_values
            if 'reference_values_kwargs' not in kwargs:
                kwargs['reference_values_kwargs'] = {'marker': 'o',
                                                     'color': 'red'}

        # call the main plotting routine from arviz
        axs = az.plot_pair(inference_data, var_names=var_names, marginals=True,
                           kind=kind, figsize=figsize, textsize=textsize,
                           **kwargs)

        # by default, the y-axis of the first and last marginal plot have ticks,
        # tick-labels and axis-labels that are not meaningful to show on the
        # y-axis; hence, we remove them here
        for i in [0, -1]:
            axs[i, i].yaxis.set_ticks_position('none')
            axs[i, i].yaxis.set_ticklabels([])
            axs[i, i].yaxis.set_visible(False)

        # adds a reference value in each marginal plot; for some reason this is
        # not done by arviz.pair_plot when passing 'reference_values'
        if 'reference_values' in kwargs:
            reference_values_kwargs = None
            if 'reference_values_kwargs' in kwargs:
                reference_values_kwargs = kwargs['reference_values_kwargs']
            ref_value_list = [*kwargs['reference_values'].values()]
            for i, prm_value in enumerate(ref_value_list):
                axs[i, i].scatter(prm_value, 0, label='true value',
                                  **reference_values_kwargs, edgecolor='black')

        if plot_priors:

            # add the prior-pdfs to the marginal subplots
            prm_names = problem.get_theta_names(tex=False)
            for i, prm_name in enumerate(prm_names):
                x = None
                if focus_on_posterior:
                    x_min, x_max = axs[i, i].get_xlim()
                    x = np.linspace(x_min, x_max, 200)
                # the following code adds labels to the prior and posterior plot
                # if they are represented as lines
                if axs[i, i].lines:
                    posterior_handle = [axs[i, i].lines[0]]
                    posterior_label = ['posterior']
                else:
                    # this is for the case, when the posterior is not shown as
                    # a line, but for example as a histogram etc.
                    posterior_handle, posterior_label = [], []
                problem.parameters[prm_name].prior.plot(
                    axs[i, i], problem.parameters, x=x)
                if show_legends:
                    prior_handle, prior_label =\
                        axs[i, i].get_legend_handles_labels()
                    axs[i, i].legend(posterior_handle + prior_handle,
                                     posterior_label + prior_label,
                                     loc='upper right')

            # here, the axis of the non-marginal plots are adjusted to the new
            # axis ranges
            if not focus_on_posterior:
                n = len(prm_names)
                for i in range(n):
                    x_min, x_max = axs[i, i].get_xlim()
                    for j in range(i + 1, n):
                        axs[j, i].set_xlim((x_min, x_max))
                    for j in range(0, i):
                        axs[i, j].set_ylim((x_min, x_max))
        else:

            # the following code adds legends to the marginal plots for the case
            # where no priors are supposed to be plotted
            if show_legends:
                prm_names = problem.get_theta_names(tex=False)
                for i, prm_name in enumerate(prm_names):
                    existing_handles, existing_labels = \
                        axs[i, i].get_legend_handles_labels()
                    if axs[i, i].lines:
                        posterior_handle = [axs[i, i].lines[0]]
                        posterior_label = ['posterior']
                    else:
                        # this is for the case, when the posterior is not shown
                        # as a line, but for example as a histogram etc.
                        posterior_handle, posterior_label = [], []
                    axs[i, i].legend(posterior_handle + existing_handles,
                                     posterior_label + existing_labels,
                                     loc='upper right')
        plt.tight_layout()
        plt.show()

    elif plot_with == 'seaborn':
        raise NotImplementedError(
            "The plot-creation with seaborn has not been implemented yet.")

    elif plot_with == 'matplotlib':
        raise NotImplementedError(
            "The plot-creation with matplotlib has not been implemented yet.")

    else:
        raise RuntimeError(
            f"Invalid 'plot_with' argument: '{plot_with}'. Available options "
            f"are currently 'arviz', 'seaborn', 'matplotlib'")

def create_posterior_plot(inference_data, problem, plot_with="arviz",
                          kind="kde", figsize=(10, 3), textsize=10, **kwargs):
    """
    Creates a posterior-plot for the given inference data using arviz.

    Parameters
    ----------
    inference_data : obj[arviz.data.inference_data.InferenceData]
        Contains the results of the sampling procedure.
    problem : obj[InferenceProblem]
        The inference problem the inference data refers to.
    plot_with : {'arviz', 'seaborn', 'matplotlib'}, optional
        Defines the python package the plot will be generated with.
    kind : str, optional
        Type of plot to display ('kde' or 'hist').
    figsize : tuple, None, optional
        Defines the size of the generated plot in the default unit. If None is
        chose, the figsize will be derived automatically.
    textsize : float, optional
        Defines the font size in the default unit.
    kwargs
        Additional keyword arguments passed to arviz' plot_posterior function.
    """

    if plot_with == "arviz":
        var_names = problem.get_theta_names(tex=True)
        az.plot_posterior(inference_data, var_names=var_names, kind=kind,
                          figsize=figsize, textsize=textsize, **kwargs)
        plt.show()

    elif plot_with == 'seaborn':
        raise NotImplementedError(
            "The plot-creation with seaborn has not been implemented yet.")

    elif plot_with == 'matplotlib':
        raise NotImplementedError(
            "The plot-creation with matplotlib has not been implemented yet.")

    else:
        raise RuntimeError(
            f"Invalid 'plot_with' argument: '{plot_with}'. Available options "
            f"are currently 'arviz', 'seaborn', 'matplotlib'")

def create_trace_plot(inference_data, problem, plot_with="arviz", kind="trace",
                      figsize=(10, 6), textsize=10, **kwargs):
    """
    Creates a trace-plot for the given inference data using arviz.

    Parameters
    ----------
    inference_data : obj[arviz.data.inference_data.InferenceData]
        Contains the results of the sampling procedure.
    problem : obj[InferenceProblem]
        The inference problem the inference data refers to.
    plot_with : {'arviz', 'seaborn', 'matplotlib'}, optional
        Defines the python package the plot will be generated with.
    kind: str, optional
        Allows to choose between plotting sampled values per iteration ("trace")
        and rank plots ("rank_bar", "rank_vlines").
    figsize : tuple, None, optional
        Defines the size of the generated plot in the default unit. If None is
        chose, the figsize will be derived automatically.
    textsize : float, optional
        Defines the font size in the default unit.
    kwargs
        Additional keyword arguments passed to arviz' plot_posterior function.
    """

    if plot_with == 'arviz':
        var_names = problem.get_theta_names(tex=True)
        az.plot_trace(inference_data, var_names=var_names, kind=kind,
                      figsize=figsize, plot_kwargs={'textsize': textsize},
                      **kwargs)
        plt.show()

    elif plot_with == 'seaborn':
        raise NotImplementedError(
            "The plot-creation with seaborn has not been implemented yet.")

    elif plot_with == 'matplotlib':
        raise NotImplementedError(
            "The plot-creation with matplotlib has not been implemented yet.")

    else:
        raise RuntimeError(
            f"Invalid 'plot_with' argument: '{plot_with}'. Available options "
            f"are currently 'arviz', 'seaborn', 'matplotlib'")



# third party imports
from scipy import stats

# local imports
from probeye.definition.prior import PriorBase


class PriorNormal(PriorBase):
    """Prior class for a normal distribution."""

    def __init__(self, ref_prm, prms_def, name):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings defining the prior's parameter names.
        name : string
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "normal distribution")

    def __call__(self, prms, method, use_ref_prm=True):
        """
        Evaluates stats.norm.<method>(x, loc, scale) or, if use_ref_prm=False
        stats.norm.<method>(loc, scale). This function is mostly used with
        method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        method : string
            The method of stats.norm to be evaluated (e.g. 'pdf' or 'logpdf').
        use_ref_prm : bool, optional
            If True stats.norm.<method>(x, loc, scale) is evaluated, hence 'x'
            must be provided in the prms dictionary. Otherwise, the evaluated
            method is stats.norm.<method>(loc, scale).

        Returns
        -------
        float
            The result of stats.norm.<method>(x, loc, scale) or of
            stats.norm.<method>(loc, scale).
        """
        fun = getattr(stats.norm, method)
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, loc=loc, scale=scale)
        else:
            return fun(loc=loc, scale=scale)

    def generate_samples(self, prms, size, seed=None):
        """
        Randomly draws samples from this prior distribution. This method is used
        to create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        size : int
            Number of samples to generate.
        seed : int or None, optional
            Used for the random state of the random number generation.

        Returns
        -------
        numpy.ndarray
            The generate samples.
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        return stats.norm.rvs(loc=loc, scale=scale, size=size,
                              random_state=seed)


class PriorLognormal(PriorBase):
    """Prior class for a log-normal distribution."""

    def __init__(self, ref_prm, prms_def, name):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings defining the prior's parameter names.
        name : string
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "log-normal distribution")

    def __call__(self, prms, method, shape=1, use_ref_prm=True):
        """
        Evaluates stats.lognorm.<method>(x, loc, scale) or, if use_ref_prm=False
        stats.lognorm.<method>(loc, scale). This function is mostly used with
        method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        method : string
            The method of stats.lognorm to be evaluated (e.g. 'pdf', 'logpdf').
        shape : float or int
            Scipy uses this shape parameter, which is not considered as a prior
            parameter here. So, it is set to 1, which results in the standard
            version of the lognormal distribution.
        use_ref_prm : bool, optional
            If True stats.norm.<method>(x, loc, scale) is evaluated, hence 'x'
            must be provided in the prms dictionary. Otherwise, the evaluated
            method is stats.norm.<method>(loc, scale).

        Returns
        -------
        float
            The result of stats.lognorm.<method>(x, loc, scale) or of
            stats.lognorm.<method>(loc, scale).
        """
        fun = getattr(stats.lognorm, method)
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, shape, loc=loc, scale=scale)
        else:
            return fun(shape, loc=loc, scale=scale)

    def generate_samples(self, prms, size, seed=None, shape=1):
        """
        Randomly draws samples from this prior distribution. This method is used
        to create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        size : int
            Number of samples to generate.
        seed : int or None, optional
            Used for the random state of the random number generation.
        shape : float or int, optional
            Scipy uses a shape parameter. For the common lognormal distribution
            this shape parameter is one.

        Returns
        -------
        numpy.ndarray
            The generate samples.
        """
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]

        return stats.lognorm.rvs(shape, loc=loc, scale=scale, size=size,
                                 random_state=seed)


class PriorUniform(PriorBase):
    """Prior class for a uniform distribution."""

    def __init__(self, ref_prm, prms_def, name):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings defining the prior's parameter names.
        name : string
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "uniform distribution")

    def __call__(self, prms, method, use_ref_prm=True):
        """
        Evaluates stats.uniform.<method>(x, loc, scale) or, if use_ref_prm=False
        stats.uniform.<method>(loc, scale). This function is mostly used with
        method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        method : string
            The method of stats.uniform to be evaluated (e.g. 'pdf', 'logpdf').
        use_ref_prm : bool, optional
            If True stats.norm.<method>(x, loc, scale) is evaluated, hence 'x'
            must be provided in the prms dictionary. Otherwise, the evaluated
            method is stats.norm.<method>(loc, scale).

        Returns
        -------
        float
            The result of stats.uniform.<method>(x, loc, scale) or of
            stats.uniform.<method>(loc, scale).
        """
        fun = getattr(stats.uniform, method)
        low = prms[f"low_{self.ref_prm}"]
        high = prms[f"high_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, loc=low, scale=high-low)
        else:
            return fun(loc=low, scale=high - low)

    def generate_samples(self, prms, size, seed=None):
        """
        Randomly draws samples from this prior distribution. This method is used
        to create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        size : int
            Number of samples to generate.
        seed : int or None, optional
            Used for the random state of the random number generation.

        Returns
        -------
        numpy.ndarray
            The generate samples.
        """
        low = prms[f"low_{self.ref_prm}"]
        high = prms[f"high_{self.ref_prm}"]
        return stats.uniform.rvs(loc=low, scale=high-low, size=size,
                                 random_state=seed)


class PriorWeibull(PriorBase):
    """Prior class for a three-parameter Weibull distribution."""

    def __init__(self, ref_prm, prms_def, name):
        """
        Parameters
        ----------
        ref_prm : string
            The name of the latent parameter the prior refers to.
        prms_def : list
            A list of strings defining the prior's parameter names.
        name : string
            Defining the priors name.
        """
        super().__init__(ref_prm, prms_def, name, "Weibull distribution")

    def __call__(self, prms, method, use_ref_prm=True):
        """
        Evaluates stats.weibull_min.<method>(x, loc, scale) or, if use_ref_prm=
        False, stats.weibull_min.<method>(loc, scale). This function is mostly
        used with method='logpdf' during the sampling procedure.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        method : string
            The method of stats.weibull_min to be evaluated (e.g. 'pdf',
            'logpdf', etc.).
        use_ref_prm : bool, optional
            If True stats.weibull_min.<method>(x, loc, scale) is evaluated,
            hence 'x' must be provided in the prms dictionary. Otherwise, the
            evaluated method is weibull_min.norm.<method>(loc, scale).

        Returns
        -------
        float
            The result of stats.weibull_min.<method>(x, loc, scale) or of
            stats.weibull_min.<method>(loc, scale).
        """
        fun = getattr(stats.weibull_min, method)
        shape = prms[f"shape_{self.ref_prm}"]
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        if use_ref_prm:
            x = prms[self.ref_prm]
            return fun(x, shape, loc=loc, scale=scale)
        else:
            return fun(shape, loc=loc, scale=scale)

    def generate_samples(self, prms, size, seed=None):
        """
        Randomly draws samples from this prior distribution. This method is used
        to create initial samples for MCMC-based algorithms.

        Parameters
        ----------
        prms : dict
            Contains the prior's parameters as keys and their values as values.
        size : int
            Number of samples to generate.
        seed : int or None, optional
            Used for the random state of the random number generation.

        Returns
        -------
        numpy.ndarray
            The generate samples.
        """
        shape = prms[f"shape_{self.ref_prm}"]
        loc = prms[f"loc_{self.ref_prm}"]
        scale = prms[f"scale_{self.ref_prm}"]
        return stats.weibull_min.rvs(shape, loc=loc, scale=scale, size=size,
                                     random_state=seed)

def translate_prior(prior_template, prior_classes=None):
    """
    Translate a given instance of PriorBase (which is essentially just a
    description of the prior without compute-methods) to a specific prior object
    which does contain compute-methods (e.g. compute the logpdf).

    Parameters
    ----------
    prior_template : obj[PriorBase]
        An instance of PriorBase which contains basic information on the
        prior but no computing-methods.
    prior_classes : dict, None
        If None, the prior classes of defined in this file are used to translate
        the prior_template. If other user-defined priors should be used, they
        can be provided via this argument, by providing a dictionary with the
        prior_type as key and the custom prior class as value.

    Returns
    -------
    prior_object : obj[PriorBase]
        An instance of a specific prior class with computing capabilities.
        Examples for such classes are given above in this file.
    """

    # check the prior_classes argument; it either must be None, or of type dict
    if type(prior_classes) is not dict:
        if prior_classes is None:
            prior_classes = {'normal': PriorNormal,
                             'lognormal': PriorLognormal,
                             'uniform': PriorUniform,
                             'weibull': PriorWeibull}
        else:
            # in this case prior_classes is not None, and not of type dict
            raise TypeError(
                f"Custom prior_classes must be given as a dictionary. However, "
                "you provided an input of type {type(prior_classes)}.")

    # prepare the corresponding prior object; the following translation is
    # necessary, because prms_def must be given in form of a list, but was
    # already translated to a dictionary when instantiating the PriorBase
    # objects; hence prior_template.prms_def is a dictionary
    prms_def = [{key: value} for key, value in prior_template.prms_def.items()
                if key != prior_template.ref_prm]
    prior_object = prior_classes[prior_template.prior_type](
        prior_template.ref_prm, prms_def, prior_template.name)

    return prior_object



# standard library
import logging
import unittest

# third party imports
import numpy as np

# local imports
from probeye.definition.forward_model import ForwardModelBase
from probeye.definition.sensor import Sensor
from probeye.definition.inference_problem import InferenceProblem
from probeye.definition.noise_model import NormalNoiseModel
from probeye.inference.emcee_.solver import EmceeSolver


class TestProblem(unittest.TestCase):

    def test_emcee_solver(self):

        np.random.seed(6174)

        # define the forward model
        class LinRe(ForwardModelBase):
            def __call__(self, inp):
                x = inp['x']
                a = inp['a']
                b = inp['b']
                return {'y': a * x + b}

        # set up the problem
        problem = InferenceProblem("Linear regression")
        problem.add_parameter(
            'a', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        problem.add_parameter(
            'b', 'model', prior=('normal', {'loc': 0, 'scale': 1}))
        problem.add_parameter(
            'sigma', 'noise', prior=('uniform',  {'low': 0.1, 'high': 1}))
        problem.add_forward_model(
            "LinRe", LinRe(['a', 'b'], [Sensor("x")], [Sensor("y")]))
        problem.add_noise_model(
            NormalNoiseModel({'sigma': 'std'}, sensors=Sensor('y')))

        # generate and add some simple test data
        n_tests, a_true, b_true, sigma_true = 5000, 0.3, -0.2, 0.1
        x_test = np.linspace(0.0, 1.0, n_tests)
        y_true = a_true * x_test + b_true
        y_test = np.random.normal(loc=y_true, scale=sigma_true)
        problem.add_experiment(f'Tests', fwd_model_name="LinRe",
                               sensor_values={'x': x_test, 'y': y_test})

        # run the emcee solver with deactivated output
        logging.root.disabled = True
        emcee_solver = EmceeSolver(problem, show_progress=False, seed=6174)
        _ = emcee_solver.run_mcmc(n_walkers=20, n_steps=200)
        # summary = run_emcee_postprocessing(problem, emcee_sampler,
        #                                    show_progress=True)
        # sample_means = summary['mean']
        # for mean, mean_true\
        #         in zip(sample_means, [a_true, b_true, sigma_true]):
        #     self.assertAlmostEqual(mean, mean_true, delta=0.01)

if __name__ == "__main__":
    unittest.main()
